import logging
import mimetypes

from django.contrib.staticfiles import finders
from twisted.internet import defer, reactor
from twisted.web import resource
from twisted.web.server import Site
from twisted.web.static import File
from twisted.web.wsgi import WSGIResource


from trosnoth.server.wsgi import application
from trosnoth.utils.twist import WeakLoopingCall
from trosnoth.web.site import PageSet, ServerState, Resources


log = logging.getLogger(__name__)

pages = PageSet()


@pages.addPage('scoreboard')
@defer.inlineCallbacks
def homepipe(state, request):
    sendData = state.getInitialEvents()
    request.write(sendData)
    while True:
        s = yield state.waitForEvent()
        if s is not None:
            request.write(s)


class WebServer(ServerState):
    def __init__(self, authFactory, serverPort):
        ServerState.__init__(self, pages)
        self.authFactory = authFactory
        self.serverPort = serverPort
        self.nextEventListeners = []
        self._loop = WeakLoopingCall(self, 'keepEventPipeAlive')
        self._loop.start(5, False)

        self.gameServer = None
        self.authFactory.onPrimaryGameChanged.addListener(
                self.gameServerChanged)

    def gameServerChanged(self, gameServer):
        if self.gameServer:
            world = self.gameServer.game.world
            world.onPlayerAdded.removeListener(self.playerCountChanged)
            world.onPlayerRemoved.removeListener(self.playerCountChanged)
            world.onGameStateChanged.removeListener(self.gameStateChanged)
            world.onTeamScoreChanged.removeListener(self.teamScoreChanged)
        self.gameServer = gameServer
        if gameServer:
            world = gameServer.game.world
            world.onPlayerAdded.addListener(self.playerCountChanged)
            world.onPlayerRemoved.addListener(self.playerCountChanged)
            world.onGameStateChanged.addListener(self.gameStateChanged)
            world.onTeamScoreChanged.addListener(self.teamScoreChanged)

        self.transmitEvent(self.getInitialEvents())

    def gameInProgress(self):
        game = self.gameServer.game
        return game.getGameState() in ('InProgress', 'Starting', 'Ended')

    def getPlayerCount(self):
        game = self.gameServer.game
        return len([p for p in game.world.players if not p.bot])

    def playerCountChanged(self, *args, **kwargs):
        if not self.gameInProgress():
            playerCount = self.getPlayerCount()
            playersString = '1 player' if playerCount == 1 else '%d players' % (
                    playerCount,)
            self.transmitEvent('message("%s in lobby.");' % (playersString,))

    def gameStateChanged(self, *args, **kwargs):
        self.transmitEvent(self.getInitialEvents())

    def getInitialEvents(self):
        if self.gameServer is None:
            return 'hideScoreboard();message("No running games on server.");\n'

        if not self.gameInProgress():
            playerCount = self.getPlayerCount()
            return 'hideScoreboard();message("%d players in lobby.");\n' % (
                    playerCount,)

        return 'hideMessage();%s\n' % (self.getScoreMessage(),)

    def teamScoreChanged(self, *args, **kwargs):
        self.transmitEvent(self.getScoreMessage())

    def getScoreMessage(self):
        world = self.gameServer.game.world
        blueTeam, redTeam = world.teams[:2]
        blueZones = blueTeam.numZonesOwned
        redZones = redTeam.numZonesOwned
        totalZones = len(world.zones)
        neutralZones = totalZones - blueZones - redZones

        return 'score(%d,%d,%d,%d,%d);' % (
            blueTeam.orbScore,
            redTeam.orbScore,
            blueZones,
            neutralZones,
            redZones,
        )

    def waitForEvent(self):
        d = defer.Deferred()
        self.nextEventListeners.append(d)
        return d

    def transmitEvent(self, jsCommand):
        listeners = self.nextEventListeners
        self.nextEventListeners = []
        for d in listeners:
            d.callback(jsCommand + '\n')

    def keepEventPipeAlive(self):
        '''
        To make sure that a reverse proxy doesn't close connections due to
        inactivity.
        '''
        self.transmitEvent('')


class WSGIRoot(resource.Resource):
    def __init__(self, wsgi, *args, **kwargs):
        resource.Resource.__init__(self)
        self.wsgi = wsgi

    def getChild(self, child, request):
        request.prepath.pop()
        request.postpath.insert(0, child)
        return self.wsgi

    def render(self, request):
        return self.wsgi.render(request)


class DjangoStatic(resource.Resource):
    isLeaf = True

    def render_GET(self, request):
        localPath = finders.find('/'.join(request.postpath))
        if not localPath:
            return File.childNotFound.render(request)

        return File(localPath).render_GET(request)


def startWebServer(authFactory, port):
    mimetypes.types_map.update({
        '.trosrepl': 'application/octet-stream',
    })

    root = WSGIRoot(
        WSGIResource(reactor, reactor.getThreadPool(), application))
    root.putChild('static', DjangoStatic())
    root.putChild('poll', Resources(WebServer(authFactory, None)))
    factory = Site(root)
    reactor.listenTCP(port, factory)



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

'''prompt.py - contains the definition for a prompt that could be used for
any type of input.'''

import logging

import pygame
import trosnoth.gui.framework.framework as framework
from trosnoth.utils.event import Event
from trosnoth.gui import keyboard

log = logging.getLogger('inputbox')

def intValidator(char):
    return char >= '0' and char <= '9'

class InputBox(framework.Element):
    pxFromLeft = 2
    pxFromRight = 5
    pxToSkipWhenCursorGoesToTheRight = 90
    cursorFlashTime = 0.6
    def __init__(self, app, area, initValue='', validator=None, font=None,
            colour=(255,255,255), maxLength=None, onClick=None, onEnter=None,
            onTab=None, onEsc=None, onEdit=None):
        super(InputBox, self).__init__(app)

        self.area = area
        self.onClick = Event()
        if onClick is not None:
            self.onClick.addListener(onClick)
        self.onEnter = Event()
        if onEnter is not None:
            self.onEnter.addListener(onEnter)
        self.onTab = Event()
        if onTab is not None:
            self.onTab.addListener(onTab)
        self.onEsc = Event()
        if onEsc is not None:
            self.onEsc.addListener(onEsc)
        self.onEdit = Event()
        if onEdit is not None:
            self.onEdit.addListener(onEdit)

        if validator:
            self.validator = validator
        else:
            self.validator = lambda x: True


        if font:
            self.font = font
        else:
            self.font = self.app.screenManager.fonts.defaultTextBoxFont

        self.value = initValue
        self.maxLength = maxLength
        self.fontColour = (0,0,0)
        self.backColour = colour
        self.cursorColour = (255, 0, 0)
        self._cursorVisible = True
        self._timeToFlash = self.cursorFlashTime

        # Calculate the white part of the input area
        self.cursorIndex = len(self.value)
        self.offset = 0

    def clear(self):
        self.setValue('')

    def setValue(self, value):
        self.value = value
        self.offset = 0
        self.cursorIndex = len(self.value)

    def getValue(self):
        return self.value

    def setFont(self, font):
        self.font = font

    def setValidator(self, validator):
        self.validator = validator

    def setMaxLength(self, length):
        self.maxLength = length

    def _getSize(self):
        return self._getRect().size

    def _getCursorImage(self):
        s = pygame.Surface((2, min(int(self.font.getHeight(self.app)),
                self._getSize()[1] * 0.8)))
        s.fill(self.cursorColour)
        return s

    def _getRect(self):
        return self.area.getRect(self.app)

    def _getPt(self):
        return self._getRect().topleft

    def setFontColour(self, fontColour):
        self.fontColour = fontColour

    def setBackColour(self, backColour):
        self.backColour = backColour

    def setCursorColour(self, cursorColour):
        self.cursorColour = cursorColour

    def gotFocus(self):
        self._cursorVisible = True
        self._timeToFlash = self.cursorFlashTime

    def _renderText(self):
        '''
        Provided so that PasswordBox can override it.
        '''
        return self.font.render(self.app, self.value, True, self.fontColour,
                self.backColour)

    def draw(self, surface):
        rect = self._getRect()
        size = rect.size
        pos = rect.topleft
        # Fill the input area with the specified colour
        surface.fill(self.backColour, rect)

        # Put what's currently inputted into the input area
        inputText = self._renderText()

        # Adjust offset
        cursorPos = self._getCursorPos()
        if cursorPos < self.offset:
            self.offset = max(0, self.offset -
                    self.pxToSkipWhenCursorGoesToTheRight)
        else:
            minOffset = cursorPos - size[0] + self.pxFromRight + self.pxFromLeft
            if self.offset < minOffset:
                self.offset = minOffset

        text_rect = inputText.get_rect()
        text_rect.centery = rect.centery
        diff = (text_rect.height - rect.height) / 2
        # Find the currently-displaying text (based on where the cursor is):
        area = pygame.Rect(self.offset, diff, size[0] - self.pxFromRight,
                rect.height)

        # Put the text on the screen
        surface.blit(inputText, (pos[0] + self.pxFromLeft, rect.top), area)

        # Add the cursor where it is.
        if self.hasFocus and self._cursorVisible:
            cs = self._getCursorImage()
            cursor_rect = cs.get_rect()
            cursor_rect.centery = rect.centery
            surface.blit(cs, (pos[0] + self._getCursorPos() - self.offset,
                    cursor_rect.top))

    def _getCursorPos(self):
        return self._getFontSize(self.value[:self.cursorIndex])[0] + 1

    def _getFontSize(self, text):
        '''
        Provided so that PasswordBox can override it.
        '''
        return self.font.size(self.app, text)

    # Get the index of the position clicked
    def __getCursorIndex(self, clickOffset):
        totalOffset = clickOffset + self.offset + self.pxFromLeft
        i = 1
        fontOffset = 0
        last = 0
        while fontOffset < totalOffset and i <= len(self.value):
            last = fontOffset
            fontOffset = self._getFontSize(self.value[:i])[0]
            i += 1
        if (fontOffset - totalOffset) <= (totalOffset - last):
            return i - 1
        else:
            return i - 2
        return i

    def processEvent(self, event):
        '''Processes the key press. If we use the event, return nothing.
        If we do not use it,
        return the event so something else can use it.'''

        if self.hasFocus and event.type == pygame.KEYDOWN:
            self._cursorVisible = True
            self._timeToFlash = self.cursorFlashTime
            if event.key in (pygame.K_RETURN, pygame.K_KP_ENTER):
                self.onEnter.execute(self)
                return None
            if event.key == pygame.K_ESCAPE:
                self.onEsc.execute(self)
                return None
            if event.key == pygame.K_TAB:
                self.onTab.execute(self)
                return None

            # Delete the letter behind the cursor
            if event.key == pygame.K_BACKSPACE:
                # Make sure there's something in the string behind the cursor
                # first, don't want to kill what's not there
                if self.cursorIndex > 0:
                    self.value = (self.value[:self.cursorIndex - 1] +
                            self.value[self.cursorIndex:])
                    self.cursorIndex -= 1
                    self.onEdit.execute(self)

            elif event.key == pygame.K_DELETE:
                # Make sure there's something in the string in front of the
                # cursor first, don't want to kill what's not there
                if self.cursorIndex < len(self.value):
                    self.value = (self.value[:self.cursorIndex] +
                            self.value[self.cursorIndex + 1:])
                    self.onEdit.execute(self)

            elif event.key == pygame.K_LEFT:
                if self.cursorIndex > 0:
                    self.cursorIndex -= 1

            elif event.key == pygame.K_RIGHT:
                if self.cursorIndex < len(self.value):
                    self.cursorIndex += 1

            elif event.key == pygame.K_END:
                self.cursorIndex = len(self.value)

            elif event.key == pygame.K_HOME:
                self.cursorIndex = 0
                self.offset = 0

            # Add the character to our string
            elif event.unicode == u'':
                return event
            else:
                # Validate the input.
                if not self.validator(event.unicode):
                    return event

                # Check the maxLength
                if self.maxLength is not None:
                    if len(self.value) >= self.maxLength:
                        return event

                # Add the typed letter to the string
                self.value = (self.value[:self.cursorIndex] + event.unicode +
                        self.value[self.cursorIndex:])
                self.cursorIndex += len(event.unicode)
                self.onEdit.execute(self)

        else:
            rect = self._getRect()

            if (event.type == pygame.MOUSEBUTTONDOWN and
                    rect.collidepoint(event.pos) and event.button == 1):

                self.onClick.execute(self)
                xOffset = event.pos[0] - rect.left
                self.cursorIndex = self.__getCursorIndex(xOffset)
                self._timeToFlash = self.cursorFlashTime
                self._cursorVisible = True
            else:
                # It's not a keydown event. Pass it on.
                return event

    def tick(self, deltaT):
        self._timeToFlash -= deltaT
        while self._timeToFlash < 0:
            self._timeToFlash += self.cursorFlashTime
            self._cursorVisible = not self._cursorVisible

class KeycodeBox(framework.Element):
    pxFromLeft = 2
    pxFromRight = 5
    def __init__(self, app, area, initValue=None, font=None,
            colour=(255,255,255), focusColour=(192, 192, 255)):
        super(KeycodeBox, self).__init__(app)

        self.area = area
        self.onClick = Event()
        self.onChange = Event()

        if font:
            self.font = font
        else:
            self.font = self.app.screenManager.fonts.defaultTextBoxFont

        self.value = initValue
        self.fontColour = (0,0,0)
        self.realBackColour = self._backColour = colour
        self.focusColour = focusColour

    def _getBackColour(self):
        return self._backColour
    def _setBackColour(self, colour):
        if self.realBackColour == self._backColour:
            self.realBackColour = colour
        self._backColour = colour
    backColour = property(_getBackColour, _setBackColour)

    def draw(self, surface):
        rect = self._getRect()
        size = rect.size
        pos = rect.topleft

        # Fill the input area with the specified colour
        surface.fill(self.realBackColour, rect)

        # Put what's currently inputted into the input area
        if self.value is None:
            name = ''
        else:
            name = keyboard.shortcutName(self.value)

        inputText = self.font.render(self.app, name, True, self.fontColour,
                        self.realBackColour)

        text_rect = inputText.get_rect()

        diff = (text_rect.height - rect.height) / 2
        area = pygame.Rect(0, diff, size[0] - self.pxFromRight, rect.height)

        # Put the text on the screen
        text_rect.centery = rect.centery
        surface.blit(inputText, (pos[0]+self.pxFromLeft, rect.top), area)

    def processEvent(self, event):
        '''Processes the key press. If we use the event, return nothing.
        If we do not use it,
        return the event so something else can use it.'''

        if self.hasFocus and event.type == pygame.KEYDOWN:
            # Catch a single keystroke.
            self.value = event.key
            self.hasFocus = False
            self.lostFocus()
            self.onChange.execute(self)
        else:
            rect = self._getRect()

            if (event.type == pygame.MOUSEBUTTONDOWN and
                    rect.collidepoint(event.pos)):
                self.onClick.execute(self)
            else:
                # It's not a keydown event. Pass it on.
                return event

    def _getSize(self):
        return self._getRect().size

    def _getRect(self):
        return self.area.getRect(self.app)

    def _getPt(self):
        return self._getRect().topleft

    def gotFocus(self):
        self.realBackColour = self.focusColour

    def lostFocus(self):
        self.realBackColour = self._backColour

class PasswordBox(InputBox):
    def _renderText(self):
        val = '*' * len(self.value)
        return self.font.render(self.app, val, True, self.fontColour,
                self.backColour)

    def _getFontSize(self, text):
        text = '*' * len(text)
        return self.font.size(self.app, text)




import logging
import pygame

from trosnoth.gui.common import (
    Location, FullScreenAttachedPoint, defaultAnchor, SizedImage,
    ScaledScreenSize, TextImage, RelativeLocation)
from trosnoth.utils.event import Event
from trosnoth.gui import colours

from framework import Element

log = logging.getLogger(__name__)


class SolidRect(Element):
    '''
    Solid rectangle.
    '''
    def __init__(self, app, colour, alpha, region, border=None):
        super(SolidRect, self).__init__(app)
        self.image = None
        self.colour = colour
        self.border = border
        self.alpha = alpha
        self.region = region

        app.screenManager.onResize.addListener(self.refresh)

    def refresh(self):
        self.image = None

    def draw(self, screen):
        r = self.region.getRect(self.app)
        if self.image is None or self.image.get_rect() != r:
            self.image = pygame.Surface(r.size)
            self.image.fill(self.colour)
            if self.alpha is not None:
                self.image.set_alpha(self.alpha)

        screen.blit(self.image, r.topleft)
        if self.border:
            pygame.draw.rect(screen, self.border, r, 1)


class PictureElement(Element):
    '''Displays a picture at a specified screen location.

    @param pos: an instance of trosnoth.ui.common.Location
    '''
    def __init__(self, app, image, pos=None):
        super(PictureElement, self).__init__(app)
        self.setImage(image)
        if pos is None:
            pos = Location(FullScreenAttachedPoint((0, 0), 'center'), 'center')
        self.pos = pos

    def setImage(self, image):
        self.image = image

    def setPos(self, pos):
        self.pos = pos

    def draw(self, screen):
        if hasattr(self.image, 'getImage'):
            surface = self.image.getImage(self.app)
        else:
            surface = self.image

        rect = pygame.Rect(surface.get_rect())
        if hasattr(self.pos, 'apply'):
            self.pos.apply(self.app, rect)
        else:
            pos = self.pos
            setattr(rect, defaultAnchor, pos)

        screen.blit(surface, rect.topleft)


class SizedPicture(PictureElement):
    '''
    Display a picture scaled to the specified position.
    @param  pos         should be an instance of trosnoth.ui.common.Location
    '''
    def __init__(self, app, surface, pos, size):
        image = SizedImage(surface, size)
        super(SizedPicture, self).__init__(app, image, pos)


# Convenience wrapper
class Backdrop(PictureElement):
    '''Display a picture scaled to the biggest 4:3 image that will fit in
    the window, with black filling the rest of the window.'''
    def __init__(self, app, name, colour):
        image = SizedImage(name, ScaledScreenSize(), False)
        self.colour = colour
        super(Backdrop, self).__init__(app, image)

    def draw(self, screen):
        if hasattr(self.image, 'getImage'):
            surface = self.image.getImage(self.app)
        else:
            surface = self.image

        rect = pygame.Rect(surface.get_rect())
        if hasattr(self.pos, 'apply'):
            self.pos.apply(self.app, rect)
        else:
            pos = self.pos
            setattr(rect, defaultAnchor, pos)

        sRect = screen.get_rect()
        if 0 < rect.left:
            screen.fill(self.colour, rect=pygame.Rect(
                0, 0, rect.left, rect.bottom))
        if 0 < rect.top:
            screen.fill(self.colour, rect=pygame.Rect(
                rect.left, 0, sRect.width - rect.left, rect.top))
        if rect.right < sRect.right:
            screen.fill(self.colour, rect=pygame.Rect(
                rect.right, rect.top, sRect.width - rect.right,
                sRect.height - rect.top))
        if rect.bottom < sRect.bottom:
            screen.fill(self.colour, rect=pygame.Rect(
                0, rect.bottom, rect.right, sRect.height - rect.bottom))

        screen.blit(surface, rect.topleft)


class TextElement(Element):
    '''Shows text at a specified screen location.
    @param pos: should be an instance of trosnoth.ui.common.Location'''

    def __init__(
            self, app, text, font, pos, colour=(0, 128, 0),
            anchor='topleft', shadow=False, backColour=None,
            antialias=True):
        super(TextElement, self).__init__(app)

        self.pos = pos
        self.anchor = anchor
        self.image = TextImage(text, font, colour, backColour, antialias)
        self.__shadow = None
        self.setShadow(shadow)
        self._surface = None
        self.rect = None

        app.screenManager.onResize.addListener(self.appResized)

    def appResized(self):
        self.rect = None
        self._surface = None

    def _getImage(self):
        if self._surface is None:
            self.image.refresh()
            self._surface = self.image.getImage(self.app)
        return self._surface

    def getRect(self, app):
        assert app is self.app
        return self._getRect()

    def _getRect(self):
        if self.rect is not None:
            return self.rect

        rect = self._getImage().get_rect()
        pos = self.pos
        if hasattr(pos, 'apply'):
            pos.apply(self.app, rect)
        else:
            setattr(rect, defaultAnchor, pos)

        self.rect = rect
        return rect

    def getText(self):
        return self.image.text

    def setColour(self, colour):
        self.image.colour = colour
        self._surface = None

    def setText(self, text):
        if self.image.text != text:
            self._surface = None
            self.image.text = text
            if self.__shadow is not None:
                self.__shadow.setText(text)

    def setFont(self, font):
        if self.image.font != font:
            self.image.font = font
            if self.__shadow is not None:
                self.__shadow.setFont(font)
            self._surface = None

    def setPos(self, pos):
        if self.pos != pos:
            self.pos = pos
            if self.__shadow is not None:
                self.__shadow.setPos(self.__getShadowPos())
            self._surface = None
            self.rect = None

    def __getShadowPos(self):
        x = self._shadowOffset
        pos = RelativeLocation(self.pos, (x, x))
        return pos

    def setShadow(self, shadow, offset=None):
        if shadow:
            if self.__shadow is None:
                if offset is None:
                    height = self.image.font.getHeight(self.app)
                    self._shadowOffset = (height / 45) + 1
                else:
                    self._shadowOffset = offset

                self.__shadow = TextElement(
                    self.app, self.image.text,
                    self.image.font, self.__getShadowPos(),
                    colours.shadowDefault, self.anchor, False)
                self._surface = None
        else:
            if self.__shadow is not None:
                self.__shadow = None
                self._surface = None

    def setShadowOffset(self, offset):
        if self.__shadow is not None:
            self.setShadow(False)
            self.setShadow(True, offset)

    def setShadowColour(self, colour):
        if self.__shadow is not None:
            self.__shadow.setColour(colour)
            self._surface = None

    def draw(self, screen):
        if self.__shadow is not None:
            assert self.__shadow.__shadow is None
            self.__shadow.draw(screen)

        rect = self._getRect()

        # Adjust text position based on anchor.
        image = self._getImage()

        rectPos = getattr(rect, self.anchor)
        imagePos = getattr(image.get_rect(), self.anchor)
        pos = (rectPos[0]-imagePos[0], rectPos[1]-imagePos[1])

        screen.blit(image, pos)


class HoverButton(Element):
    '''A button which changes image when the mouse hovers over it.
    @param  pos     should be a trosnoth.ui.common.Location instance
    '''
    def __init__(
            self, app, pos, stdImage, hvrImage, hotkey=None, onClick=None):
        super(HoverButton, self).__init__(app)

        self.stdImage = stdImage
        self.hvrImage = hvrImage
        self.pos = pos
        self.onClick = Event()
        self.hotkey = hotkey

        self.mouseOver = False

        if onClick is not None:
            self.onClick.addListener(onClick)

    def _getSurfaceToBlit(self):
        img = self._getImageToUse()
        if hasattr(img, 'getImage'):
            return img.getImage(self.app)
        else:
            return img

    def _getImageToUse(self):
        if self.mouseOver:
            return self.hvrImage
        else:
            return self.stdImage

    def _getSize(self):
        img = self._getImageToUse()
        if hasattr(img, 'getSize'):
            return img.getSize(self.app)
        else:
            # pygame.surface.Surface
            return img.get_size()

    def _getPt(self):
        if hasattr(self.pos, 'apply'):
            rect = pygame.Rect((0, 0), self._getSize())
            self.pos.apply(self.app, rect)
            return rect.topleft
        else:
            return self.pos

    def _getRect(self):
        if hasattr(self.pos, 'apply'):
            rect = pygame.Rect((0, 0), self._getSize())
            self.pos.apply(self.app, rect)
            return rect
        return pygame.Rect(self.pos, self._getSize())

    def processEvent(self, event):
        rect = self._getRect()
        # Passive events.
        if event.type == pygame.MOUSEMOTION:
            self.mouseOver = rect.collidepoint(event.pos)

        # Active events.
        if (event.type == pygame.MOUSEBUTTONDOWN and event.button == 1 and
                rect.collidepoint(event.pos)):
            self.onClick.execute(self)
        elif event.type == pygame.KEYDOWN and event.key == self.hotkey:
            self.onClick.execute(self)
        else:
            return event
        return None

    def draw(self, screen):
        # Draw the button.
        screen.blit(self._getSurfaceToBlit(), self._getPt())


class TextButton(HoverButton):
    'A HoverButton which has text rather than images.'
    def __init__(
            self, app, pos, text, font, stdColour, hvrColour,
            hotkey=None, backColour=None, onClick=None):
        self.stdImage = TextImage(text, font, stdColour, backColour)
        self.hvrImage = TextImage(text, font, hvrColour, backColour)

        super(TextButton, self).__init__(
            app, pos, self.stdImage, self.hvrImage, hotkey, onClick)

        self.app.screenManager.onResize.addListener(self.appResized)

    def appResized(self):
        # The font may be scaled, so refresh the images
        self.stdImage.refresh()
        self.hvrImage.refresh()

    def setText(self, text):
        self.stdImage.text = text
        self.hvrImage.text = text

    def setFont(self, font):
        self.stdImage.setFont(font)
        self.hvrImage.setFont(font)

    def setColour(self, colour):
        self.stdImage.setColour(colour)

    def setHoverColour(self, colour):
        self.hvrImage.setColour(colour)

    def setBackColour(self, colour):
        self.hvrImage.setBackColour(colour)
        self.stdImage.setBackColour(colour)



import logging

import pygame
from twisted.internet import defer
from twisted.internet.error import ConnectError
from twisted.protocols import amp

from trosnoth.gui.common import (Region, Canvas, Location, ScaledSize)
from trosnoth.gui.framework import framework
from trosnoth.gui.framework.dialogbox import DialogBox, OkBox
from trosnoth.gui.framework.elements import (TextElement, SolidRect, TextButton)
from trosnoth.network.lobby import (Lobby, AuthenticationCancelled,
        IncorrectServerVersion)
from trosnoth.network import authcommands
from trosnoth.network.client import ConnectionFailed
from trosnoth.utils.event import Event
from trosnoth.trosnothgui.pregame.authServerLoginBox import (PasswordGUI,
        PasswordGUIError)


log = logging.getLogger('playscreen')

class PlayAuthScreen(framework.CompoundElement):
    passwordGUIFactory = PasswordGUI

    def __init__(self, app, onSucceed=None, onFail=None):
        super(PlayAuthScreen, self).__init__(app)
        self.onSucceed = Event(onSucceed)
        self.onFail = Event(onFail)
        self.lobby = None
        self.badServers = set()

        if app.displaySettings.alphaOverlays:
            alpha = 192
        else:
            alpha = None
        bg = SolidRect(app, app.theme.colours.playMenu, alpha,
                Region(centre=Canvas(512, 384), size=Canvas(924, 500)))

        colour = app.theme.colours.mainMenuColour
        font = app.screenManager.fonts.consoleFont
        self.logBox = LogBox(app, Region(size=Canvas(900, 425),
                midtop=Canvas(512, 146)), colour, font)

        font = app.screenManager.fonts.bigMenuFont
        cancel = TextButton(app, Location(Canvas(512, 624), 'midbottom'),
                'Cancel', font, app.theme.colours.secondMenuColour,
                app.theme.colours.white, onClick=self.cancel)
        self.cancelled = False
        self.elements = [bg, self.logBox, cancel]
        self._passwordGetter = None

    @property
    def passwordGetter(self):
        if self._passwordGetter is None:
            self._passwordGetter = self.passwordGUIFactory(self.app)
        return self._passwordGetter

    @defer.inlineCallbacks
    def begin(self, servers, canHost=True):
        self.cancelled = False
        self.passwordGetter.cancelled = False
        self.badServers = set()
        if self.lobby is None:
            self.lobby = Lobby(self.app)

        # Removes the third item (http) from the tuple since we don't care about
        # it.
        servers = [ (server[:2] if isinstance(server, tuple) else server)
                for server in servers ]

        for server in servers:
            if self.cancelled:
                break

            if server == 'self':
                if self.app.server is not None:
                    self.onSucceed.execute()
                    self.app.interface.connectToLocalServer()
                    return

            elif isinstance(server, tuple):
                if server in self.badServers:
                    continue

                self.logBox.log('Requesting games from %s:%d...' % server)
                connected = yield self.attemptServerConnect(
                        self.lobby.getGames, server)
                if connected:
                    return

            elif server == 'others':
                for server in servers:
                    if (server in self.badServers or not isinstance(server,
                            tuple)):
                        continue

                    self.logBox.log('Asking %s:%d about other games...' %
                            server)
                    connected = yield self.attemptServerConnect(
                            self.lobby.getOtherGames, server)
                    if connected:
                        return

            elif server == 'lan':
                self.logBox.log('Asking local network for other games...')
                games = yield self.lobby.getMulticastGames()
                for game in games:
                    joinSuccessful = yield self.attemptJoinGame(game)
                    if joinSuccessful:
                        return

            elif server == 'create':
                for server in servers:
                    if server in self.badServers or not isinstance(server,
                            tuple):
                        continue
                    self.logBox.log('Asking to create game on %s...' %
                            (server[0],))

                    joinSuccessful = yield self.attemptCreateGame(server)
                    if joinSuccessful:
                        return

        if canHost:
            if not self.cancelled:
                result = yield HostGameQuery(self.app).run()

                if not result:
                    self.onFail.execute()
                    return

                self.app.startServer(2, 1)

                # Notify remaining auth servers of this game.
                for server in servers:
                    if server in self.badServers or not isinstance(server,
                            tuple):
                        continue
                    self.logBox.log('Registering game with %s...' %
                            (server[0],))
                    try:
                       result = yield self.lobby.registerGame(server,
                               self.app.server)
                    except ConnectError:
                        self.logBox.log('Unable to connect.')
                    if not result:
                        self.logBox.log('Registration failed.')

                self.onSucceed.execute()
                self.app.interface.connectToLocalServer()
        else:
            if not self.cancelled:
                box = OkBox(self.app, ScaledSize(450, 150), 'Trosnoth',
                        'Connection unsuccessful.')
                box.onClose.addListener(self.onFail.execute)
                box.show()

    @defer.inlineCallbacks
    def attemptServerConnect(self, getGamesList, server):
        '''
        Attempts to connect to a game on the server, returned by
        getGamesList(server).
        '''
        try:
           games = yield getGamesList(server)
        except ConnectError:
            self.logBox.log('Unable to connect.')
            self.badServers.add(server)
        except amp.UnknownRemoteError:
            self.logBox.log('Error on remote server.')
        except amp.RemoteAmpError, e:
            self.logBox.log('Error on remote server.')
            log.exception(str(e))
        except IOError, e:
            self.logBox.log('Error connecting to remote server.')
            log.exception(str(e))
            self.badServers.add(server)
        else:
            if len(games) == 0:
                self.logBox.log('No running games.')

            for game in games:
                joinSuccessful = yield self.attemptJoinGame(game)
                if joinSuccessful:
                    defer.returnValue(True)
                    return

        defer.returnValue(False)

    @defer.inlineCallbacks
    def attemptCreateGame(self, server):
        '''
        Attempts to create a game on the given server.
        '''
        try:
           result = yield self.lobby.startGame(server,
                   self.passwordGetter)
        except (ConnectError, ConnectionFailed):
            self.logBox.log('Unable to connect.')
            self.badServers.add(server)
        except authcommands.CannotCreateGame:
            self.logBox.log('Server will not create game.')
        except IncorrectServerVersion:
            self.logBox.log('Wrong server version to create games.')
        except AuthenticationCancelled:
            pass
        except authcommands.NotAuthenticated:
            self.logBox.log('Authentication failure.')
        except authcommands.GameDoesNotExist:
            self.logBox.log('Error connecting to newly-created game.')
        except amp.UnknownRemoteError:
            self.logBox.log('Error on remote server.')
        except amp.RemoteAmpError, e:
            self.logBox.log('Error on remote server.')
            log.exception(str(e))
        except IOError, e:
            self.logBox.log('Error connecting to remote server.')
            log.exception(str(e))
        else:
            self.logBox.log('Game created and joined.')
            self._joined(result)
            defer.returnValue(True)
            return

        defer.returnValue(False)

    @defer.inlineCallbacks
    def attemptJoinGame(self, game):
        '''
        Attempts to join the given game.
        '''
        try:
            self.logBox.log('Found game: joining.')
            result = yield game.join(self.passwordGetter)
        except authcommands.GameDoesNotExist:
            pass
        except AuthenticationCancelled:
            pass
        except authcommands.NotAuthenticated:
            self.logBox.log('Authentication failure.')
        except amp.UnknownRemoteError:
            self.logBox.log('Error on remote server.')
        except amp.RemoteAmpError, e:
            self.logBox.log('Error on remote server.')
            log.exception(str(e))
        except ConnectionFailed:
            self.logBox.log('Could not connect.')
        except IOError, e:
            self.logBox.log('Error connecting to remote server.')
            log.exception(str(e))
        else:
            self._joined(result)
            defer.returnValue(True)
            return

        defer.returnValue(False)

    def _joined(self, result):
        self.onSucceed.execute()
        self.app.interface.connectedToGame(*result)

    def cancel(self, element):
        self.cancelled = True
        self.passwordGetter.cancelled = True
        self.onFail.execute()

    def processEvent(self, event):
        if event.type == pygame.KEYDOWN and event.key == pygame.K_ESCAPE:
            self.onFail.execute()
        else:
            return super(PlayAuthScreen, self).processEvent(event)

class LogBox(framework.Element):
    '''
    Draws a canvas and adds text to the bottom of it, scrolling the canvas
    upwards.
    '''
    def __init__(self, app, region, colour, font):
        super(LogBox, self).__init__(app)
        self.image = None
        self.region = region
        self.colour = colour
        self.font = font

    def draw(self, screen):
        r = self.region.getRect(self.app)
        if self.image is not None:
            screen.blit(self.image, r.topleft)

    def log(self, text):
        t = self.font.render(self.app, text, False, self.colour,
                (255, 255, 255))
        t.set_colorkey((255, 255, 255))
        r = self.region.getRect(self.app)
        img = pygame.Surface(r.size)
        img.fill((255, 255, 255))
        img.set_colorkey((255, 255, 255))
        h = t.get_rect().height
        r.topleft = (0, h)
        if self.image is not None:
            img.blit(self.image, (0, 0), r)
        img.blit(t, (0, r.height - h))
        self.image = img


class HostGameQuery(DialogBox):
    def __init__(self, app):
        size = Canvas(384, 150)
        DialogBox.__init__(self, app, size, 'Host game?')
        self._deferred = None

        font = app.screenManager.fonts.defaultTextBoxFont
        btnColour = app.theme.colours.dialogButtonColour
        highlightColour = app.theme.colours.black
        labelColour = app.theme.colours.dialogBoxTextColour
        btnFont = app.screenManager.fonts.bigMenuFont

        self.elements = [
            TextElement(app, 'No games found. Host a game?', font,
                Location(self.Relative(0.5, 0.4), 'center'), labelColour),

            TextButton(app,
                Location(self.Relative(0.3, 0.85), 'center'),
                'Yes', btnFont, btnColour, highlightColour,
                onClick=self.yesClicked),
            TextButton(app,
                Location(self.Relative(0.7, 0.85), 'center'),
                'No', btnFont, btnColour, highlightColour,
                onClick=self.noClicked),
        ]

    def processEvent(self, event):
        if event.type == pygame.KEYDOWN:
            if event.key in (pygame.K_KP_ENTER, pygame.K_RETURN):
                self.yesClicked()
                return None
            elif event.key == pygame.K_ESCAPE:
                self.noClicked()
                return None
        else:
            return DialogBox.processEvent(self, event)

    def yesClicked(self, element=None):
        self.close()
        self._deferred.callback(True)

    def noClicked(self, element=None):
        self.close()
        self._deferred.callback(False)

    def run(self):
        if self.showing:
            raise PasswordGUIError('HostGameQuery already showing')
        self.show()
        result = self._deferred = defer.Deferred()
        return result




# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2011 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from twisted.internet import defer

from trosnoth.run.authserver import (AuthenticationFactory, AuthManager,
        AuthenticatedUser)
import trosnoth.run.authserver

class CampAuthManager(AuthManager):
    def authenticateUser(self, username, password):
        username = username.lower()
        if password == username.encode('rot13'):
            user = AuthenticatedUser(self, username)
            user.seen()
            return defer.succeed(user)
        return super(CampAuthManager, self).authenticateUser(username, password)

class CampAuthenticationFactory(AuthenticationFactory):
    authManagerClass = CampAuthManager

class main(trosnoth.run.authserver.main):
    authFactory = CampAuthenticationFactory

if __name__ == '__main__':
    main()




import pygame

from trosnoth.const import MAP_TO_SCREEN_SCALE

def blitPart(surface, source, dest, part):
    '''
    Performs a blit, positioning the source on the surface as for
    surface.blit(source, dest), but only blits the subrect part. part is a rect
    relative to the top-left corner of the source image.
    '''
    surface.blit(source, (dest[0] + part.left, dest[1] + part.top), part)


def viewRectToMap(focus, area):
    pos = screenToMapPos(area.topleft, focus, area)
    size = (
        int(area.width / MAP_TO_SCREEN_SCALE + 0.5),
        int(area.height / MAP_TO_SCREEN_SCALE + 0.5))
    return pygame.Rect(pos, size)


def zonePosToScreen(pt, focus, area):
    return mapPosToScreen((pt[0] - 1024, pt[1] - 384), focus, area)


def mapPosToScreen(pt, focus, area):
    return (int((pt[0] - focus[0]) * MAP_TO_SCREEN_SCALE + area.centerx + 0.5),
            int((pt[1] - focus[1]) * MAP_TO_SCREEN_SCALE + area.centery + 0.5))


def screenToMapPos(pt, focus, area):
    return (int((pt[0] - area.centerx) / MAP_TO_SCREEN_SCALE + 0.5) + focus[0],
            int((pt[1] - area.centery) / MAP_TO_SCREEN_SCALE + 0.5) + focus[1])


# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from trosnoth.model.map import MapLayout

def getBlocksInRect(world, rect):
    '''
    Returns an iterator of the blocks in the given rect.
    '''
    i, j0 = MapLayout.getMapBlockIndices(rect.left, rect.top)
    j0 = max(0, j0)
    i = max(0, i)

    while i < len(world.zoneBlocks):
        row = world.zoneBlocks[i]
        j = j0
        while j < len(row):
            block = row[j]
            blockLeft, blockTop = block.defn.pos
            if blockTop >= rect.bottom:
                return
            if blockLeft >= rect.right:
                break
            yield block
            j += 1
        i += 1

def getZonesInRect(world, rect):
    '''
    This will yield all zones which are in map blocks in the given rect.
    Note that sometimes a zone will be yielded that is not quite in the rect, if
    one of the zone's map blocks is in the rect.
    '''
    touchedZones = set()
    for block in getBlocksInRect(world, rect):
        for zone in block.getZones():
            if zone not in touchedZones:
                touchedZones.add(zone)
                yield zone




# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('trosnoth', '0003_auto_20160207_0026'),
    ]

    operations = [
        migrations.AddField(
            model_name='trosnothserversettings',
            name='iceEnabled',
            field=models.BooleanField(default=False),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='trosnothserversettings',
            name='iceHost',
            field=models.TextField(default=b'127.0.0.1'),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='trosnothserversettings',
            name='icePort',
            field=models.IntegerField(default=6502),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='trosnothserversettings',
            name='iceProxyStringOverride',
            field=models.TextField(default=b''),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='trosnothserversettings',
            name='iceSecret',
            field=models.TextField(default=b''),
            preserve_default=True,
        ),
    ]



import logging
import os
import time

import pygame
from Tkinter import Tk
from tkFileDialog import askopenfilename
from twisted.internet import tksupport

from trosnoth.data import getPath, user, makeDirs
from trosnoth.gui import browser
from trosnoth.gui.common import (
    Location, ScaledLocation, ScaledArea, AttachedPoint,
)
from trosnoth.gamerecording.gamerecorder import (
    RecordedGame, RecordedGameException, gameDir, gameExt, recordedGameVersion,
    replayDir,
)
from trosnoth.gui.framework.elements import (
    TextElement, TextButton, SizedPicture,
)
import trosnoth.gui.framework.framework as framework
from trosnoth.gui.framework.listbox import ListBox
from trosnoth.gui.framework.tab import Tab
from trosnoth.gui.framework.tabContainer import TabContainer, TabSize
from trosnoth.utils.event import Event

log = logging.getLogger(__name__)


class SavedGameMenu(framework.CompoundElement):
    def __init__(self, app, onCancel=None, onReplay=None):
        super(SavedGameMenu, self).__init__(app)

        area = ScaledArea(50, 140, 924, 570)
        bg = pygame.Surface((924, 500))
        bg.fill(app.theme.colours.replayMenu)
        if app.displaySettings.alphaOverlays:
            bg.set_alpha(192)

        font = self.app.screenManager.fonts.ampleMenuFont

        self.tabContainer = TabContainer(
            self.app, area, font, app.theme.colours.replayTabBorder)
        bp = SizedPicture(
            app, bg, Location(
                AttachedPoint((0, 0), self.tabContainer._getTabRect)),
            TabSize(self.tabContainer))

        self.replayTab = SavedGameTab(
            app, self.tabContainer, onCancel=onCancel, onReplay=onReplay)
        self.tabContainer.addTab(self.replayTab)

        self.elements = [bp, self.tabContainer]


class SavedGameTab(Tab):
    def __init__(self, app, tabContainer, onCancel=None, onReplay=None):
        super(SavedGameTab, self).__init__(app, 'Saved Games')
        self.app = app
        self.tabContainer = tabContainer
        self.onCancel = Event(onCancel)
        self.onReplay = Event(onReplay)

        font = self.app.screenManager.fonts.ampleMenuFont
        smallFont = self.app.screenManager.fonts.menuFont
        colours = app.theme.colours

        # Static text
        self.staticText = [
            TextElement(
                self.app, 'server details:', font,
                ScaledLocation(960, 200, 'topright'), colours.headingColour),
            TextElement(
                self.app, 'date and time:', font,
                ScaledLocation(960, 370, 'topright'), colours.headingColour),
            TextElement(
                self.app, 'replay:', font,
                ScaledLocation(620, 550, 'topleft'), colours.headingColour),
            TextElement(
                self.app, 'stats:', font,
                ScaledLocation(620, 605, 'topleft'), colours.headingColour)]

        # Dynamic text
        self.listHeaderText = TextElement(
            self.app, 'available game files:', font, ScaledLocation(65, 200),
            colours.headingColour)
        self.noFiles1Text = TextElement(
            self.app, '', font, ScaledLocation(65, 260), colours.noGamesColour)
        self.noFiles2Text = TextElement(
            self.app, '', font, ScaledLocation(65, 310), colours.noGamesColour)
        self.serverNameText = TextElement(
            self.app, '', smallFont, ScaledLocation(960, 255, 'topright'),
            colours.startButton)
        self.serverDetailsText = TextElement(
            self.app, '', smallFont, ScaledLocation(960, 295, 'topright'),
            colours.startButton)
        self.dateText = TextElement(
            self.app, '', smallFont, ScaledLocation(960, 425, 'topright'),
            colours.startButton)
        self.lengthText = TextElement(
            self.app, '', smallFont, ScaledLocation(960, 465, 'topright'),
            colours.startButton)
        self.noReplayText = TextElement(
            self.app, '', smallFont, ScaledLocation(960, 550, 'topright'),
            colours.noGamesColour)
        self.noStatsText = TextElement(
            self.app, '', smallFont, ScaledLocation(960, 605, 'topright'),
            colours.noGamesColour)

        self.dynamicText = [
            self.listHeaderText, self.noFiles1Text, self.noFiles2Text,
            self.serverNameText, self.serverDetailsText, self.dateText,
            self.lengthText, self.noReplayText, self.noStatsText]

        # Text buttons
        self.watchButton = TextButton(
            self.app, ScaledLocation(960, 550, 'topright'), '', font,
            colours.secondMenuColour, colours.white)
        self.watchButton.onClick.addListener(self.watchReplay)

        self.statsButton = TextButton(
            self.app, ScaledLocation(960, 605, 'topright'), '', font,
            colours.secondMenuColour, colours.white)
        self.statsButton.onClick.addListener(self.viewStats)

        self.refreshButton = TextButton(
            self.app, ScaledLocation(620, 665, 'topleft'), 'refresh', font,
            colours.secondMenuColour, colours.white)
        self.refreshButton.onClick.addListener(self.populateList)

        self.cancelButton = TextButton(
            self.app, ScaledLocation(960, 665, 'topright'), 'cancel', font,
            colours.secondMenuColour, colours.white)
        self.cancelButton.onClick.addListener(self._cancel)

        self.loadFileButton = TextButton(
            self.app, ScaledLocation(960, 190, 'bottomright'), 'load file...',
            font, colours.mainMenuColour, colours.mainMenuHighlight)
        self.loadFileButton.onClick.addListener(self.showOpenDialog)

        self.buttons = [
            self.watchButton, self.statsButton, self.refreshButton,
            self.cancelButton, self.loadFileButton]

        # Replay list
        self.gameList = ListBox(
            self.app, ScaledArea(65, 255, 500, 450), [], smallFont,
            colours.listboxButtons)
        self.gameList.onValueChanged.addListener(self.updateSidebar)

        # Combine the elements
        self.elementsFiles = (
            self.staticText + self.dynamicText + self.buttons
            + [self.gameList])
        self.elementsNoFiles = self.dynamicText + self.buttons

        # Populate the list of replays
        self.populateList()

    def _cancel(self, sender):
        self.onCancel.execute()

    def showOpenDialog(self, sender):
        root = Tk()
        root.withdraw()
        tksupport.install(root)
        filename = askopenfilename(
            defaultextension='.trosrepl',
            filetypes=[
                ('Trosnoth replay', '*.trosrepl'),
            ],
            initialdir=getPath(user, replayDir),
            title='Select replay',
        )
        if filename:
            self.onReplay.execute(filename)

    def populateList(self, sender=None):

        # Clear out the sidebar
        for item in self.dynamicText:
            item.setText('')
        self.listHeaderText.setText('available game files:')
        self.gameList.index = -1
        self.elements = self.elementsFiles[:]

        # Get a list of files with the name '*.tros'
        logDir = getPath(user, gameDir)
        makeDirs(logDir)
        fileList = []

        for fname in os.listdir(logDir):
            if os.path.splitext(fname)[1] == gameExt:
                fileList.append(fname)

        # Assume all files are valid for now
        validFiles = fileList[:]

        self.gameInfo = {}
        oldFound = False

        for fname in fileList:
            try:
                game = RecordedGame(os.path.join(logDir, fname))
            except RecordedGameException:
                validFiles.remove(fname)
                continue
            except:
                log.warning('invalid file: %s', fname)
                continue
            else:
                if game.recordedGameVersion != recordedGameVersion:
                    validFiles.remove(fname)
                    oldFound = True

            self.gameInfo[os.path.splitext(fname)[0]] = game

        # Sort the games with most recent first.
        items = [(v.unixTimestamp, n) for n, v in self.gameInfo.iteritems()]
        items.sort(reverse=True)
        items = [n for v, n in items]
        self.gameList.setItems(items)

        if len(self.gameInfo) == 0:
            self.elements = self.elementsNoFiles[:]
            self.listHeaderText.setText('0 available game files:')
            if oldFound:
                self.noFiles1Text.setText('Some games were found from')
                self.noFiles2Text.setText('previous Trosnoth versions')
            else:
                self.noFiles1Text.setText('You have not yet run any')
                self.noFiles2Text.setText('games on this computer')
        else:
            self.gameList.setIndex(0)
            self.updateSidebar(0)
            if len(self.gameInfo) == 1:
                self.listHeaderText.setText('1 available game file:')
            else:
                self.listHeaderText.setText(
                    '{} available game files:'.format(len(self.gameInfo)))

    def updateSidebar(self, listID):
        # Update the details on the sidebar
        displayName = self.gameList.getItem(listID)

        # Server title
        self.serverNameText.setText(self.gameInfo[displayName].alias)

        # Map size
        self.serverDetailsText.setText('Map size: {} x {}'.format(
            self.gameInfo[displayName].halfMapWidth,
            self.gameInfo[displayName].mapHeight))

        # Date and time of match
        datePython = map(int, self.gameInfo[displayName].dateTime.split(','))
        dateString = time.strftime('%a %d/%m/%y, %H:%M', datePython)
        self.dateText.setText(dateString)

        # Length of match
        dateUnix = time.mktime(datePython)
        if self.gameInfo[displayName].wasFinished():
            lastUnix = self.gameInfo[displayName].gameFinishedTimestamp

            lengthSeconds = int(lastUnix - dateUnix)
            lengthMinutes, lengthSeconds = divmod(lengthSeconds, 60)

            secPlural = ('s', '')[lengthSeconds == 1]
            minPlural = ('s', '')[lengthMinutes == 1]
            if lengthMinutes == 0:
                lengthString = '{} second{}'.format(lengthSeconds, secPlural)
            else:
                lengthString = '{} min{}, {} sec{}'.format(
                    lengthMinutes, minPlural, lengthSeconds, secPlural)

            self.lengthText.setText(lengthString)
        else:
            self.lengthText.setText('')

        # Enable the replay button
        if (self.gameInfo[displayName].replayFilename is not None and
                os.path.exists(self.gameInfo[displayName].replayFilename)):
            self.watchButton.setText('watch')
            self.noReplayText.setText('')
        else:
            self.watchButton.setText('')
            self.noReplayText.setText('unavailable')

        # Enabled the stats button
        if (self.gameInfo[displayName].statsFilename is not None and
                os.path.exists(self.gameInfo[displayName].statsFilename)):
            self.statsButton.setText('view')
            self.noStatsText.setText('')
        else:
            self.statsButton.setText('')
            self.noStatsText.setText('unavailable')

    def watchReplay(self, sender):
        '''Watch replay button was clicked.'''
        # Try to create a replay server.
        self.onReplay.execute(
            self.gameInfo[self.gameList.getItem()].replayFilename)

    def viewStats(self, sender=None):
        '''View stats button was clicked.'''
        game = self.gameInfo[self.gameList.getItem()]

        self.htmlPath = game.generateHtmlFile()
        browser.openPage(self.app, self.htmlPath)

    def draw(self, surface):
        super(SavedGameTab, self).draw(surface)

        rect = self.tabContainer._getTabRect()
        verLineX = rect.left + (rect.width * 0.6)
        horLineY = rect.top + (rect.height * 0.68)

        colour = self.app.theme.colours.replayTabBorder

        pygame.draw.line(
            surface, colour, (verLineX, rect.top), (verLineX, rect.bottom),
            self.tabContainer._getBorderWidth())
        pygame.draw.line(
            surface, colour, (verLineX, horLineY), (rect.right, horLineY),
            self.tabContainer._getBorderWidth())



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from trosnoth.gui.framework import framework
from trosnoth.gui.framework.elements import PictureElement, TextElement
from trosnoth.gui.common import Location, AttachedPoint
from trosnoth.utils.event import Event
import pygame


class RadioButtonGroup(framework.CompoundElement):
    def __init__(self, app):
        super(RadioButtonGroup, self).__init__(app)
        self.elements = []
        self.selectedIndex = -1
        self.onSelectionChanged = Event()

    def add(self, radioButton):
        self.elements.append(radioButton)
        radioButton._joinGroup(self)

    def getSelectedValue(self):
        if self.selectedIndex == -1:
            return None
        else:
            return self.elements[self.selectedIndex].value

    def selected(self, radioButton):
        assert radioButton in self.elements, ("Cannot select a radiobutton "
                "that isn't in this group")

        newIndex = self.elements.index(radioButton)
        if newIndex == self.selectedIndex:
            return

        if self.selectedIndex != -1:
            self.elements[self.selectedIndex].deselected()
        self.selectedIndex = newIndex
        radioButton.selected()
        self.onSelectionChanged.execute(self.selectedIndex)


class ImageRadioButton(framework.CompoundElement):

    class IRB_background(framework.Element):
        def __init__(self, app, imageRadioButton):
            super(ImageRadioButton.IRB_background, self).__init__(app)
            self.imageRadioButton = imageRadioButton
            self.mouseOver = False

        def draw(self, surface):
            if self.imageRadioButton._selected:
                colour = self.app.theme.colours.radioSelected
            elif self.mouseOver:
                colour = self.app.theme.colours.radioMouseover
            else:
                colour = self.app.theme.colours.radioUnselected
            surface.fill(colour, self.imageRadioButton._getRect())

        def processEvent(self, event):
            rect = self.imageRadioButton._getRect()
            # Passive events.
            if event.type == pygame.MOUSEMOTION:
                self.mouseOver = rect.collidepoint(event.pos)

            # Active events.
            if (event.type == pygame.MOUSEBUTTONDOWN and
                    rect.collidepoint(event.pos)):
                self.imageRadioButton._clicked()
                return None
            else:
                return event


    def __init__(self, app, text, area, image):
        super(ImageRadioButton, self).__init__(app)
        textLoc = Location(AttachedPoint((0,0), self._getRect, 'midbottom'),
                'midbottom')
        font = app.screenManager.fonts.smallNoteFont
        self.text = TextElement(app, text, font, textLoc, (0,0,0))
        picLoc = Location(AttachedPoint((0,0), self._getRect, 'midtop'),
                'midtop')
        self.area = area

        self.value = None

        self._selected = False
        self.group = None

        self.bg = self.IRB_background(app, self)

        self.elements = [self.bg, self.text]

        if image is not None:
            self.image = PictureElement(app, image, picLoc)
            self.elements.insert(1, self.image)


    def _beforeSelected(self):
        return True

    def _clicked(self):
        assert self.group, ("Image Radio Buttons are required to belong to "
                "a RadioButtonGroup")
        if self._beforeSelected():
            self.group.selected(self)

    def _getRect(self):
        return self.area.getRect(self.app)

    # Should only be called by the parent radio button group
    def _joinGroup(self, group):
        self.group = group

    def selected(self):
        self._selected = True

    def deselected(self):
        self._selected = False

    def draw(self, surface):
        super(ImageRadioButton, self).draw(surface)
        pygame.draw.rect(surface, self.app.theme.colours.black,
                         self._getRect(), 2)



# coding: utf-8
from __future__ import division

from django.contrib.auth.models import User
from django.contrib.staticfiles import finders
from django.http import Http404
from django.shortcuts import render

from trosnoth.gamerecording.achievementlist import availableAchievements
from trosnoth.model.upgrades import upgradeOfType
from .models import (
    TrosnothServerSettings, TrosnothUser, GameRecord, PlayerKills,
    UpgradesUsedInGameRecord,
)

POINT_VALUES = {
    'kills': 10,
    'deaths': 1,
    'zoneTags': 20,
    'accuracy': 20,
    'starsUsed': 3,
    'zoneAssists': 5,
}

def index(request):
    context = {
        'settings': TrosnothServerSettings.get(),
    }
    return render(request, 'trosnoth/index.html', context)


def userProfile(request, userId, nick=None):
    try:
        user = TrosnothUser.fromUser(pk=userId)
    except User.DoesNotExist:
        raise Http404('User not found')

    unlocked = []
    locked = []
    for a in user.achievementprogress_set.all():
        try:
            name, description = availableAchievements.getAchievementDetails(
                a.achievementId)
        except KeyError:
            if not a.unlocked:
                continue
            name = a.achievementId
            description = (
                'This achievement does not exist in this version of Trosnoth')

        if finders.find(
                'trosnoth/achievements/{}.png'.format(a.achievementId)):
            imageId = a.achievementId
        else:
            imageId = 'default'

        info = {
            'name': name,
            'description': description,
            'imageId': imageId,
        }
        if a.unlocked:
            unlocked.append(info)
        else:
            locked.append(info)
    unlocked.sort(key=lambda a: a['name'])
    locked.sort(key=lambda a: a['name'])

    context = {
        'settings': TrosnothServerSettings.get(),
        'trosnothUser': user,
        'unlocked': unlocked,
        'locked': locked,
    }
    return render(request, 'trosnoth/user.html', context)


def userList(request):
    context = {
        'settings': TrosnothServerSettings.get(),
        'users': TrosnothUser.objects.order_by('-lastSeen'),
    }
    return render(request, 'trosnoth/userlist.html', context)


def viewGame(request, gameId):
    game = GameRecord.objects.get(pk=gameId)

    data = []

    for player in game.gameplayer_set.all():
        entry = {
            'player': player,
            'nick': player.user.nick if player.user else player.botName,
            'accuracy': (100.0 * player.shotsHit / player.shotsFired
                ) if player.shotsFired else 0.,
            'score': 0,
            'kdr': u'{:2.2f}'.format(player.kills / player.deaths
                ) if player.deaths else u'∞',
            'adr': u'{:2.2f}'.format(player.timeAlive / player.timeDead
                ) if player.timeDead else u'∞',
        }

        for stat, value in POINT_VALUES.iteritems():
            if stat in entry:
                value = entry[stat]
            else:
                value = getattr(player, stat)
            entry['score'] += value

        data.append(entry)

    data.sort(key=(lambda entry: entry['score']), reverse=True)

    i = 1
    j = 0
    for entry in data:
        entry['index'] = j
        if entry['player'].bot:
            entry['rank'] = 'B'
        else:
            entry['rank'] = str(i)
            i += 1
        j += 1

    killData = {}
    for pkr in PlayerKills.objects.filter(killee__game=game):
        killData[pkr.killer, pkr.killee] = pkr.count

    killTable = []
    for killerEntry in data:
        killer = killerEntry['player']
        killRow = []
        maxKillCount = maxDeathCount = 0
        maxKill = maxDeath = '-'
        for killeeEntry in data:
            killee = killeeEntry['player']
            count = killData.get((killer, killee), 0)
            killRow.append(count)
            if count > maxKillCount:
                maxKillCount = count
                maxKill = '{} ({})'.format(killeeEntry['nick'], count)
            dieCount = killData.get((killee, killer), 0)
            if dieCount > maxDeathCount:
                maxDeathCount = dieCount
                maxDeath = '{} ({})'.format(killeeEntry['nick'], dieCount)
        killerEntry['maxKill'] = maxKill
        killerEntry['maxDeath'] = maxDeath

        killTable.append({
            'player': killerEntry['player'],
            'nick': killerEntry['nick'],
            'entries': killRow,
        })

    for i in xrange(len(killTable)):
        killTable[i]['entries'][i] = '-'

    otherKills = [
        killData.get((None, killeeEntry['player']), 0)
        for killeeEntry in data
    ]

    upgradeData = {}
    upgradeCodes = set()
    for ur in UpgradesUsedInGameRecord.objects.filter(gamePlayer__game=game):
        upgradeData[ur.gamePlayer, ur.upgrade] = ur.count
        upgradeCodes.add(ur.upgrade)

    if upgradeCodes:
        nameAndCode = []
        for code in upgradeCodes:
            if code in upgradeOfType:
                name = upgradeOfType[code].name
            else:
                name = '?{}?'.format(code)
            nameAndCode.append((name, code))

        nameAndCode.sort()
        upgradeList = [name for name, code in nameAndCode]
        upgradeTable = []
        for entry in data:
            entries = []
            maxUpgrade = '-'
            maxUpgradeCount = 0
            for name, code in nameAndCode:
                count = upgradeData.get((entry['player'], code), 0)
                entries.append(count)
                if count > maxUpgradeCount:
                    maxUpgrade = '{} ({})'.format(name, count)

            entry['maxUpgrade'] = maxUpgrade

            upgradeTable.append({
                'player': entry['player'],
                'nick': entry['nick'],
                'entries': entries,
            })
    else:
        upgradeList = []
        upgradeTable = []

    context = {
        'settings': TrosnothServerSettings.get(),
        'game': game,
        'playerData': data,
        'killTable': killTable,
        'otherKills': otherKills if any(otherKills) else None,
        'upgrades': upgradeList,
        'upgradeTable': upgradeTable,
    }
    return render(request, 'trosnoth/viewgame.html', context)


def gameList(request):
    context = {
        'settings': TrosnothServerSettings.get(),
        'games': GameRecord.objects.order_by('-started'),
    }
    return render(request, 'trosnoth/gamelist.html', context)



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import pygame
import UserDict

# 1. We want to be able to represent a shortcut as a string.
def shortcutName(key, modifiers=0):
    # pygame.key.name() could also work.
    try:
        name = NAMED[key]
    except KeyError:
        name = pygame.key.name(key)

    # Add the modifiers.
    modString = ''
    for kmod, modName in KMOD_NAMES:
        if modifiers & kmod:
            modString = '%s%s+' % (modString, modName)

    return '%s%s' % (modString, name)

NAMED = {
    pygame.K_BACKSPACE: 'Backspace', pygame.K_BREAK: 'Break', pygame.K_CAPSLOCK:
    'Capslock', pygame.K_CLEAR: 'Clear', pygame.K_DELETE: 'Del', pygame.K_DOWN:
    'Down', pygame.K_END: 'End', pygame.K_ESCAPE: 'Escape', pygame.K_EURO:
    'Euro', pygame.K_F1: 'F1', pygame.K_F2: 'F2', pygame.K_F3: 'F3',
    pygame.K_F4: 'F4', pygame.K_F5: 'F5', pygame.K_F6: 'F6', pygame.K_F7: 'F7',
    pygame.K_F8: 'F8', pygame.K_F9: 'F9', pygame.K_F10: 'F10', pygame.K_F11:
    'F11', pygame.K_F12: 'F12', pygame.K_F13: 'F13', pygame.K_F14: 'F14',
    pygame.K_F15: 'F15', pygame.K_FIRST: 'First', pygame.K_HELP: 'Help',
    pygame.K_HOME: 'Home', pygame.K_INSERT: 'Ins', pygame.K_LALT: 'L.Alt',
    pygame.K_LCTRL: 'L.Ctrl', pygame.K_LEFT: 'Left',
    pygame.K_LMETA: 'L.Meta', pygame.K_LSHIFT: 'L.Shift', pygame.K_LSUPER:
    'L.Super', pygame.K_MENU: 'Menu', pygame.K_MODE: 'Mode', pygame.K_NUMLOCK:
    'Numlock', pygame.K_PAGEDOWN: 'PgDn', pygame.K_PAGEUP: 'PgUp',
    pygame.K_PAUSE: 'Pause', pygame.K_POWER: 'Power', pygame.K_PRINT: 'Print',
    pygame.K_RALT: 'R.Alt', pygame.K_RCTRL: 'R.Ctrl', pygame.K_RETURN: 'Return',
    pygame.K_RIGHT: 'Right', pygame.K_RMETA: 'R.Meta', pygame.K_RSHIFT:
    'R.Shift', pygame.K_RSUPER: 'R.Super', pygame.K_SCROLLOCK: 'Scrolllock',
    pygame.K_SYSREQ: 'SysRq', pygame.K_TAB: 'Tab', pygame.K_UP: 'Up',
    pygame.K_SPACE: 'Space',

    pygame.K_a: 'a', pygame.K_b: 'b', pygame.K_c: 'c', pygame.K_d: 'd',
    pygame.K_e: 'e', pygame.K_f: 'f', pygame.K_g: 'g', pygame.K_h: 'h',
    pygame.K_i: 'i', pygame.K_j: 'j', pygame.K_k: 'k', pygame.K_l: 'l',
    pygame.K_m: 'm', pygame.K_n: 'n', pygame.K_o: 'o', pygame.K_p: 'p',
    pygame.K_q: 'q', pygame.K_r: 'r', pygame.K_s: 's', pygame.K_t: 't',
    pygame.K_u: 'u', pygame.K_v: 'v', pygame.K_w: 'w', pygame.K_x: 'x',
    pygame.K_y: 'y', pygame.K_z: 'z', pygame.K_0: '0', pygame.K_1: '1',
    pygame.K_2: '2', pygame.K_3: '3', pygame.K_4: '4', pygame.K_5: '5',
    pygame.K_6: '6', pygame.K_7: '7', pygame.K_8: '8', pygame.K_9: '9',
    pygame.K_KP0: 'keypad-0', pygame.K_KP1: 'keypad-1', pygame.K_KP2:
    'keypad-2', pygame.K_KP3: 'keypad-3', pygame.K_KP4: 'keypad-4',
    pygame.K_KP5: 'keypad-5', pygame.K_KP6: 'keypad-6', pygame.K_KP7:
    'keypad-7', pygame.K_KP8: 'keypad-8', pygame.K_KP9: 'keypad-9',
    pygame.K_KP_DIVIDE: 'keypad divide', pygame.K_KP_ENTER: 'keypad enter',
    pygame.K_KP_EQUALS: 'keypad equals', pygame.K_KP_MINUS: 'keypad minus',
    pygame.K_KP_MULTIPLY: 'keypad asterisk', pygame.K_KP_PERIOD:
    'keypad full stop', pygame.K_KP_PLUS: 'keypad plus',

    pygame.K_AMPERSAND: '&', pygame.K_ASTERISK: '*', pygame.K_AT: '@',
    pygame.K_BACKQUOTE: '`', pygame.K_BACKSLASH: '\\', pygame.K_CARET: '^',
    pygame.K_COLON: ':', pygame.K_COMMA: ',', pygame.K_DOLLAR: '$',
    pygame.K_EQUALS: '=', pygame.K_EXCLAIM: '!', pygame.K_GREATER: '>',
    pygame.K_LESS: '<', pygame.K_HASH: '#', pygame.K_LEFTBRACKET: '[',
    pygame.K_LEFTPAREN: '(', pygame.K_MINUS: '-', pygame.K_PERIOD: '.',
    pygame.K_PLUS: '+', pygame.K_QUOTE: "'", pygame.K_RIGHTBRACKET: ']',
    pygame.K_RIGHTPAREN: ')', pygame.K_SEMICOLON: ';', pygame.K_SLASH: '/',
    pygame.K_UNDERSCORE: '_'
 }

KMOD_NAMES = ((pygame.KMOD_CTRL, 'Ctrl'), (pygame.KMOD_ALT, 'Alt'),
              (pygame.KMOD_META, 'Meta'), (pygame.KMOD_SHIFT, 'Shift'))


# VirtualKeySet is a mapping from name -> default value.
class VirtualKeySet(UserDict.UserDict):
    pass

# KeyboardMapping is a mapping from key -> virtual key name.
class KeyboardMapping(UserDict.UserDict):
    def __init__(self, virtualKeys):
        self.virtualKeys = virtualKeys
        UserDict.UserDict.__init__(self, ((default, vk) for (vk, default) in
                virtualKeys.iteritems()))

    def load(self, string):
        '''
        Restores a keyboard mapping from a configuration string.
        '''
        # Reset to defaults.
        self.data = {}

        # Update from string.
        unmappedKeys = dict(self.virtualKeys)
        for record in string.split('\n'):
            if record == '':
                continue
            key, vk = record.split(':')
            self.data[int(key)] = vk
            if vk in unmappedKeys:
                del unmappedKeys[vk]

        # Fill in any unmapped keys from the defaults if possible.
        for vk, default in unmappedKeys.iteritems():
            if default not in self.data:
                self.data[default] = vk

    def save(self):
        '''
        Returns a configuration string for this keyboard mapping.
        '''
        records = ['%d:%s' % item for item in self.data.iteritems()]
        return '\n'.join(records)

    def getkey(self, action):
        '''
        Returns one key that results in the given action or raises KeyError.
        '''
        for k, v in self.data.iteritems():
            if v == action:
                return k
        raise KeyError(action)



import logging

import pygame
from twisted.internet import reactor

from trosnoth.game import LocalGame
from trosnoth.settings import (
    DisplaySettings, SoundSettings, IdentitySettings, ConnectionSettings,
)
from trosnoth.trosnothgui import interface
from trosnoth.themes import Theme
from trosnoth.manholehelper import LocalManholeHelper
from trosnoth.model.mapLayout import LayoutDatabase
from trosnoth.network.lobby import UDPMulticaster
from trosnoth.network.networkDefines import serverVersion
from trosnoth.network.server import TrosnothServerFactory

from trosnoth.gui import app

log = logging.getLogger(__name__)

GAME_TITLE = 'Trosnoth'


class Main(app.MultiWindowApplication):
    '''Instantiating the Main class will set up the game. Calling the run()
    method will run the reactor. This class handles the three steps of joining
    a game: (1) get list of clients; (2) connect to a server; (3) join the
    game.'''

    standardInterfaceFactory = interface.Interface
    singleInterfaceFactory = interface.SingleAuthInterface

    def __init__(self, serverDetails=None, showReplay=None):
        '''Initialise the game.'''
        pygame.init()

        self.serverDetails = serverDetails
        self.server = None
        self.serverInvisible = False

        self.displaySettings = DisplaySettings(self)
        self.soundSettings = SoundSettings(self)
        self.identitySettings = IdentitySettings(self)
        self.connectionSettings = ConnectionSettings(self)

        pygame.font.init()
        if self.serverDetails is not None:
            iface = self.singleInterfaceFactory
        else:
            iface = self.standardInterfaceFactory
            if showReplay:
                reactor.callWhenRunning(self.loadReplay, showReplay)

        super(Main, self).__init__(
            self.displaySettings.getSize(),
            self.displaySettings.fullScreen,
            GAME_TITLE, iface)

        # Set the master sound volume.
        self.soundSettings.apply()

        # Start listening for game requests on the lan.
        self.multicaster = UDPMulticaster(self.getGames)

    def __str__(self):
        return 'Trosnoth Main Object'

    def getConsoleLocals(self):
        result = {
            'server': self.server,
            'helper': LocalManholeHelper(
                lambda: getattr(self, 'interface', None).game),
        }
        try:
            result['game'] = self.interface.game
        except AttributeError:
            pass
        return result

    def stopping(self):
        # Shut down the server if one's running.
        if self.server is not None:
            self.server.shutdown()

        self.multicaster.stop()
        super(Main, self).stopping()

    def initialise(self):
        super(Main, self).initialise()

        # Loading the theme loads the fonts.
        self.theme = Theme(self)
        self.layoutDatabase = LayoutDatabase(pathGetter=self.theme.getPath)

    def loadReplay(self, filename):
        self.screenManager.interface.connectToReplay(filename)

    def getFontFilename(self, fontName):
        '''
        Tells the UI framework where to find the given font.
        '''
        return self.theme.getPath('fonts', fontName)

    def changeScreenSize(self, size, fullScreen):
        self.screenManager.setScreenProperties(size, fullScreen, GAME_TITLE)

    def startServer(
            self, halfMapWidth=3, mapHeight=2,
            gameDuration=0, invisibleGame=False):
        if self.server is not None and self.server.running:
            return
        game = LocalGame(
            self.layoutDatabase, halfMapWidth, mapHeight,
            duration=gameDuration * 60,
            saveReplay=True, gamePrefix='LAN')
        self.server = TrosnothServerFactory(game)
        self.serverInvisible = invisibleGame

        self.server.onShutdown.addListener(self._serverShutdown)
        self.server.startListening()

    def getGames(self):
        '''
        Called by multicast listener when a game request comes in.
        '''
        if self.server and not self.serverInvisible:
            gameInfo = {
                'version': serverVersion,
                'port': self.server.getTCPPort(),
            }
            return [gameInfo]
        return []

    def _serverShutdown(self):
        self.server.stopListening()
        self.server = None
        self.serverInvisible = False



'''mapLayout.py - takes care of initialising map blocks with obstacles and
 images.'''

import logging
import os
import random
from StringIO import StringIO

import pygame

from trosnoth import data
from trosnoth.utils.utils import hasher
from trosnoth.utils import unrepr
from trosnoth.model.map import MapLayout

from trosnoth.data import getPath, user, makeDirs
import trosnoth.data.blocks

log = logging.getLogger('mapLayout')

BLOCK_FILES = {
    'ForwardInterfaceMapBlock': 'fwd',
    'BackwardInterfaceMapBlock': 'bck',
    'TopBodyMapBlock': 'top',
    'BottomBodyMapBlock': 'btm',
}


class ChecksumError(Exception):
    pass


class LayoutDatabase(object):
    '''
    Represents a database which stores information on block layouts.
    Contains several LayoutDatastores which will be queried in their order of
    priority when a block is needed.
    '''

    def __init__(self, blocks=(), pathGetter=None):
        if pathGetter is None:
            pathGetter = self.getDefaultPath
        self.pathGetter = pathGetter
        self.datastores = []
        self._addDatastoreWithBlocks(blocks)
        self._addDatastoreByPaths(
            getPath(trosnoth.data.blocks), getPath(user, 'blocks'))
        self.downloadPath = getPath(user, 'downloadedBlocks')
        self.downloads = self._addDatastoreByPaths(self.downloadPath)

    def getDefaultPath(self, *bits):
        return data.getPath(data, *bits)

    def _addDatastoreWithBlocks(self, blocks):
        if len(blocks) == 0:
            return

        store = LayoutDatastore(pathGetter=self.pathGetter)
        self.datastores.append(store)
        for block in blocks:
            store.addLayoutAtFilename(block)

    def _addDatastoreByPaths(self, *paths):
        for path in paths:
            makeDirs(path)

        store = LayoutDatastore(pathGetter=self.pathGetter)
        self.datastores.append(store)

        # Read map blocks from files.
        for path in paths:
            filenames = os.listdir(path)

            # Go through all files in the blocks directory.
            for fn in filenames:
                # Check for files with a .block extension.
                if os.path.splitext(fn)[1] == '.block':
                    store.addLayoutAtFilename(os.path.join(path, fn))
        return store

    def addDownloadedBlock(self, key, contents, graphics):
        '''
        Adds the given block (which has been downloaded from a remote server)
        to the database.
        '''
        # First save the contents.
        while True:
            base = '%.32x' % (random.randrange(1 << 128))
            filename = os.path.join(self.downloadPath, base)
            if not os.path.exists(filename):
                break
        open(filename + '.block', 'wb').write(contents)
        open(filename + '.png', 'wb').write(graphics)

        # Add new block to datastore.
        self.downloads.addLayoutAtFilename(filename + '.block')

    def getFundamentalKey(self, key):
        '''
        If a block with the given key exists in the database, returns None,
        otherwise returns a fundamental form of the key which should be used
        when asking a remote server for a copy of the block. This is done so
        that only one request is sent for map blocks which are mirror images of
        one another.
        '''
        kind, checksum, reversed = key
        for ds in self.datastores:
            if (kind, checksum) in ds.layoutsByKey:
                return None
        return (kind, checksum, False)

    def getLayoutByKey(self, key):
        '''
        Returns a BlockLayout from its key, or None if no such layout is known.
        '''
        for ds in self.datastores:
            result = ds.getLayoutByKey(key)
            if result is not None:
                return result
        return None

    def getRandomLayout(self, kind, blocked):
        for ds in self.datastores:
            if len(ds.layouts[kind, blocked]) > 0:
                return random.choice(ds.layouts[kind, blocked])
        raise ValueError(
            'No layouts found of kind %r with blocked %r' % (kind, blocked))

    def getRandomSymmetricalLayout(self, kind, blocked):
        for ds in self.datastores:
            if len(ds.symmetricalLayouts[kind, blocked]) > 0:
                return random.choice(ds.symmetricalLayouts[kind, blocked])
        raise ValueError(
            'No symmetrical layouts found of kind %r with blocked %r' % (
                kind, blocked))

    def randomiseBlock(self, block, oppBlock):
        '''Takes a map block and gives it and the
        corresponding opposite block a layout depending on its block type
        and whether it has a barrier.
        '''
        if oppBlock is not block:
            # The block is not symmetrical.
            layout = self.getRandomLayout(block.kind, block.blocked)
            layout.applyTo(block)
            layout.mirrorLayout.applyTo(oppBlock)
        else:
            # The block is symmetrical.
            layout = self.getRandomSymmetricalLayout(block.kind, block.blocked)
            layout.applyTo(block)

    def loadMapFromFile(self, filename):
        mapDir = getPath(user, 'maps')
        filename = os.path.join(mapDir, filename)
        if not os.path.exists(filename):
            raise IOError("Map could not be loaded; it doesn't exist!")

        layoutStr = open(filename, 'r').read()

        return MapLayout.fromString(self, layoutStr)

    def generateRandomMapLayout(self, halfMapWidth, mapHeight, blockRatio=0.5):
        '''
        Generates and returns a random MapLayout object of the given dimensions
        using blocks from this layout database.
        '''
        layout = MapLayout(halfMapWidth, mapHeight)
        mapBlockList = {}
        tempBlocked = {}
        for blockList in layout.blocks:
            for block in blockList:
                tempBlocked[block] = False
                zones = block.getZones()
                for zone in zones:
                    mapBlockList.setdefault(zone, []).append(block)

        x = -1
        for blockList in layout.blocks:
            x += 1
            for i in range(len(blockList) / 2 + 1):
                block = blockList[i]
                oppBlock = blockList[-i - 1]

                if block.blocked:
                    # The block is on the edge, leave it as is
                    continue
                elif len(block.getZones()) == 0:
                    # The block is outside where the player can go
                    continue
                else:
                    # If it's a topBodyMapBlock, make it the same as the
                    # bottomBodyMapBlock just above it
                    if block.kind == 'top':
                        blockAbove = layout.blocks[x - 1][i]
                        tempBlocked[block] = tempBlocked[blockAbove]
                    elif random.random() < blockRatio:
                        tempBlocked[block] = True
                        tempBlocked[oppBlock] = True

        # At this point, many of the blocks will probably be connected to all
        # others. However, we shall now ensure that all are connected, using
        # the following method:
        #
        #
        # 1. Find the central zone of the entire map. Use this as a starting
        # point.
        #
        # 2. Create a list of connected zones. At first, the list should only
        # contain the central zone.
        #
        # 3. Create an empty list of unconnected but adjacent zones.
        #
        # 4. For each zone next to the central zone, if it is blocked, add it
        # to the list of unconnected but adjacent zones. Otherwise, add it to
        # the list of connected zones. However, ignore zones on the right-hand
        # side of the map.
        #
        # 5. Repeat step 4 for each of the zones that was just added to the
        # list of connecteds. If ever a zone is added to the list of
        # connecteds, be sure to remove it from the list of
        # unconnected-but-adjacent-zones (if it exists in that list).
        #
        # 6. Choose a random zone from the list of unconnected and adjacent
        # zones. Connect it to a random zone in the list of connecteds
        # (obviously one to which it is adjacent). Remove it from the list
        # of unconnecteds, and add it to the list of connecteds.
        #
        # 7. While the list of unconnecteds is not empty, repeat steps 4-6,
        # using this newly connected zone in place of the central zone.

        centralRow = layout.blocks[len(layout.blocks) / 2]
        centralBlock = centralRow[len(centralRow) / 2]
        centralZone = centralBlock.zone

        def adjacentZones(zone):
            '''
            yields (adjZone, connected) for each adjacent zone, where connected
            is True or False depending on whether the zones are connected or
            not.
            '''
            for zone2, blocks in zone.adjacentZones.iteritems():
                for block in blocks:
                    if tempBlocked[block]:
                        yield zone2, False
                        break
                else:
                    yield zone2, True

        def connectZones(zone1, zone2):
            for block in zone1.adjacentZones[zone2]:
                tempBlocked[block] = False

        mapWidth = (
            (layout.halfMapWidth * 2 + 1) * (
                layout.zoneBodyWidth + layout.zoneInterfaceWidth)
            + layout.zoneInterfaceWidth)

        def oppositeZone(zone):
            x, y = zone.pos
            x = mapWidth - x
            i, j = layout.getMapBlockIndices(x, y)
            return layout.blocks[i][j].getZoneAtPoint(x, y)

        visitedZones = set()
        newZones = set([centralZone])
        unreachedZones = {}     # zone -> connected zones
        while len(newZones) > 0 or len(unreachedZones) > 0:
            while len(newZones) > 0:
                curZone = newZones.pop()
                visitedZones.add(curZone)
                for newZone, connected in adjacentZones(curZone):
                    if newZone in visitedZones:
                        continue
                    if connected:
                        if newZone in unreachedZones:
                            del unreachedZones[newZone]
                        newZones.add(newZone)
                    elif newZone not in visitedZones:
                        unreachedZones.setdefault(newZone, []).append(curZone)

            if len(unreachedZones) > 0:
                zone1 = random.choice(unreachedZones.keys())
                zone2 = random.choice(unreachedZones.pop(zone1))
                connectZones(zone1, zone2)
                newZones.add(zone1)
                zone1b = oppositeZone(zone1)
                zone2b = oppositeZone(zone2)
                if zone1b != zone1 or zone2b != zone2:
                    connectZones(zone1b, zone2b)
                    if zone1b in unreachedZones:
                        del unreachedZones[zone1b]
                    newZones.add(zone1b)

        # Populate the map blocks with appropriate obstacles.

        for blockList in layout.blocks:
            for i in range(len(blockList) / 2 + 1):
                block = blockList[i]
                oppBlock = blockList[-i - 1]
                if block is oppBlock:
                    assert block.kind in ('top', 'btm')
                if not block.blocked:
                    block.blocked = tempBlocked[block]
                    if oppBlock is not None:
                        oppBlock.blocked = tempBlocked[oppBlock]
                if not block.getZones() == []:
                    # If the block contains zones, get a layout for this block.
                    self.randomiseBlock(block, oppBlock)

        return layout


class LayoutDatastore(object):
    '''Represents a database which stores information on block layouts.'''

    def __init__(self, pathGetter):
        '''(paths) - initialises the database and loads the blocks from the
        specified paths.'''
        self.pathGetter = pathGetter

        # Set up database.
        self.layouts = {}
        self.layoutsByFilename = {}
        self.layoutsByKey = {}          # Keyed by (kind, checksum)
        self.symmetricalLayouts = {}
        for b in True, False:
            for a in 'fwd', 'bck', 'top', 'btm':
                self.layouts[a, b] = []
            for a in 'top', 'btm':
                self.symmetricalLayouts[a, b] = []

    def addLayoutAtFilename(self, filepath):
        # Remember the filename.
        self.filename = os.path.split(filepath)[1]

        # Read the file and create the block
        f = open(filepath, 'rU')
        try:
            contents = f.read()
        finally:
            f.close()

        self.checksum = hasher(contents).hexdigest()

        try:
            contentDict = unrepr.unrepr(contents)
            self.addLayout(filepath, **contentDict)
        except Exception, e:
            log.warning(str(e))
            return False
        return True

    def addLayout(
            self, filepath, blockType='TopBodyMapBlock', blocked=False,
            obstacles=[], platforms=[], symmetrical=False, graphics=None):
        '''
        Registers a layout with the given parameters.
        '''
        base, ext = os.path.splitext(filepath)
        if graphics is None:
            graphicsIO = None
            try:
                graphicPath = self.pathGetter(
                    'blocks', '%s.png' % (os.path.split(base)[1],))
            except IOError:
                graphicPath = base + '.png'
            else:
                if not os.path.exists(graphicPath):
                    graphicPath = base + '.png'
        else:
            graphicPath = None
            graphicsIO = StringIO()
            graphicsIO.write(graphics.decode('base64'))

        blockType = BLOCK_FILES[blockType]

        newLayout = BlockLayout()
        newLayout.setProperties(
            graphicPath, obstacles, platforms, symmetrical, self.checksum,
            blockType, filepath, graphicPath, graphicsIO)

        # Add the layout to the database.
        self.layouts[blockType, blocked].append(newLayout)
        self.layoutsByFilename[self.filename] = newLayout
        self.layoutsByKey[(blockType, self.checksum)] = newLayout

        if symmetrical:
            self.symmetricalLayouts[blockType, blocked].append(newLayout)
        else:
            if blockType == 'fwd':
                blockType = 'bck'
            elif blockType == 'bck':
                blockType = 'fwd'

            self.layouts[blockType, blocked].append(newLayout.mirrorLayout)

    def getLayoutByKey(self, key):
        '''
        Returns a BlockLayout from its key, or None if no such layout is known.
        '''
        kind, checksum, reversed = key
        try:
            layout = self.layoutsByKey[(kind, checksum)]
        except KeyError:
            return None
        if reversed:
            return layout.mirrorLayout
        return layout


class BlockLayout(object):
    '''Represents the layout of a block. Saves the positions of all obstacles
    within the block as well as a graphic of the block.'''

    def __init__(self):
        '''() - initialises a blank block layout.'''
        self.filename = None
        self.graphics = None
        self.obstacles = []
        self.platforms = []
        self.mirrorLayout = self
        self.reversed = False
        self.kind = None
        self.graphicsFilename = None
        self.graphicsIO = None

    @property
    def forwardLayout(self):
        if self.reversed:
            return self.mirrorLayout
        else:
            return self

    @property
    def key(self):
        '''
        Returns a key by which this block layout may be uniquely identified.
        The block may be obtained from the LayoutDatabase using its
        .getLayoutByKey() method.
        '''
        return (self.kind, self.checksum, self.reversed)

    def setProperties(
            self, graphicPath, obstacles, platforms, symmetrical,
            checksum, blockType, filename, graphicsFilename, graphicsIO):
        '''(graphicPath, obstacles, symmetrical) - sets the properties of
        this block layout.

        graphicPath     A string representing the path to the graphic file
                        which describes the background of this block, or
                        None if no graphic file should be used.
        obstacles       Should be a sequence of obstacle definitions, each of
                        which should be of the form ((x1, y1), (dx, dy)).
                        Obstacles can be passed through in one direction only.
                        A solid block should be composed of obstacles defined
                        in a clockwise direction.
        platforms       Obstacles which can be fallen through.
        symmetrical     Boolean - is this block symmetrical? If set to True,
                        this block will also create a block which is the exact
                        mirror of this block.
        '''
        self.checksum = checksum
        self.kind = blockType
        self.filename = filename
        self.graphicsFilename = graphicsFilename
        self.graphicsIO = graphicsIO

        # Set up the graphic.
        if graphicsIO is None:
            if graphicPath:
                # Check that the graphic exists.
                if not os.path.exists(graphicPath):
                    raise IOError(
                        'Graphic file does not exist: %s' % (graphicPath,))
            self.graphics = FileBlockGraphics(graphicPath)
        else:
            self.graphics = StreamBlockGraphics(graphicsIO)

        # Record obstacles.
        self.obstacles = []
        for obstacle in obstacles:
            self.obstacles.append(obstacle)

        self.platforms = []
        for platform in platforms:
            self.platforms.append(tuple(platform))

        # If it's not symmetrical, create a mirror block.
        if symmetrical:
            self.mirrorLayout = self
        else:
            self.mirrorLayout = BlockLayout()
            self.mirrorLayout.reversed = True
            self.mirrorLayout.mirrorLayout = self
            self.mirrorLayout.checksum = checksum
            if graphicsIO is None:
                self.mirrorLayout.graphics = FileBlockGraphics(
                    graphicPath, True)
            else:
                self.mirrorLayout.graphics = StreamBlockGraphics(
                    graphicsIO, True)
            self.mirrorLayout.kind = self.kind
            self.mirrorLayout.filename = self.filename
            self.mirrorLayout.graphicsFilename = self.graphicsFilename
            self.mirrorLayout.graphicsIO = self.graphicsIO

            # Flip the obstacles.
            for obstacle in self.obstacles:
                mObstacle = []
                for (x, y) in obstacle:
                    mObstacle.insert(0, (-x, y))
                self.mirrorLayout.obstacles.append(tuple(mObstacle))
            for platform in self.platforms:
                x1, y1 = platform[0]
                dx = platform[1]
                x2 = x1 + dx
                self.mirrorLayout.platforms.append(((-x2, y1), dx))

    def applyTo(self, block):
        '''Applies this layout to the specified map block.'''
        block.layout = self
        block.graphics = self.graphics

        # Go through the obstacles and append them to the block's obstacles.
        for obstacle in self.obstacles:
            block.addObstacle(obstacle, self.reversed)

        for relPos, dx in self.platforms:
            block.addPlatform(relPos, dx, self.reversed)

        block.coalesceLayout()

    def getGraphicsData(self):
        if self.graphicsIO is not None:
            self.graphicsIO.seek(0)
            return self.graphicsIO.read()
        return open(self.graphicsFilename, 'rb').read()


class BaseBlockGraphics(object):
    '''Defers loading of graphic until needed.'''

    def __init__(self, reversed):
        self._miniGraphics = {}
        self._rawGraphic = None
        self._graphic = None
        self._reversed = reversed

    def getMini(self, app, scale):
        try:
            result = self._miniGraphics[scale]
        except KeyError:
            width = int(self.getGraphic(app).get_rect().width / scale + 0.5)
            height = int(self.getGraphic(app).get_rect().height / scale + 0.5)

            result = pygame.transform.smoothscale(
                self._rawGraphic, (width, height))
            self._miniGraphics[scale] = result

        return result

    def getGraphic(self, app):
        if self._graphic is not None:
            return self._graphic

        # Load the graphic.
        self._rawGraphic = pygame.image.load(self._getFile(), 'png')

        if self._reversed:
            # Flip the graphic.
            self._rawGraphic = pygame.transform.flip(
                self._rawGraphic, True, False)

        self._graphic = self._rawGraphic.convert_alpha()

        if not app.displaySettings.perPixelAlpha:
            # Create a surface that doesn't have the per-pixel alpha
            result = pygame.Surface(self._graphic.get_size())
            result.fill((127, 127, 127))
            result.blit(self._graphic, (0, 0))
            result.set_colorkey((127, 127, 127))
            self._graphic = result

        return self._graphic


class StreamBlockGraphics(BaseBlockGraphics):
    def __init__(self, stream, reversed=False):
        BaseBlockGraphics.__init__(self, reversed)
        self._stream = stream

    def _getFile(self):
        result = StringIO(self._stream.getvalue())
        return result


class FileBlockGraphics(BaseBlockGraphics):
    '''Defers loading of graphic until needed.'''

    def __init__(self, filename, reversed=False):
        BaseBlockGraphics.__init__(self, reversed)
        self._filename = filename

    def _getFile(self):
        return self._filename



import logging

import os

import pygame
from twisted.internet import reactor

from trosnoth.const import MAP_TO_SCREEN_SCALE
from trosnoth.data import getPath, user
from trosnoth.utils import unrepr
from trosnoth.utils.event import Event
import trosnoth.version

log = logging.getLogger('settings')


class SettingsObject(object):
    '''
    Base class for defining settings objects. Defines some functionality for
    loading and saving settings.
    '''

    def __init__(self, app):
        self.app = app
        self.reset()

    def reset(self):
        data = self._loadSettingsFile()

        for attr, key, default in self.attributes:
            setattr(self, attr, data.get(key, default))

    def apply(self):
        pass

    def save(self):
        # Write to file
        fn = getPath(user, self.dataFileName)
        f = open(fn, 'w')
        data = {}
        for attr, key, default in self.attributes:
            data[key] = getattr(self, attr)
        f.write(repr(data))
        f.close()

    def _getSettingsFilename(self):
        return getPath(user, self.dataFileName)

    def _loadSettingsFile(self):
        filename = self._getSettingsFilename()
        try:
            f = open(filename, 'r')
            d = unrepr.unrepr(f.read())
            if not isinstance(d, dict):
                d = {}
        except IOError:
            d = {}
        return d


class DisplaySettings(SettingsObject):
    '''
    Stores the Trosnoth display settings.
    '''
    DEFAULT_THEME = 'default'
    dataFileName = 'display'
    attributes = (
        ('size', 'size', None),
        ('fullScreen', 'fullscreen', False),
        ('_old_useAlpha', 'usealpha', True),
        ('_old_windowsTranslucent', 'windowsTranslucent', False),
        ('detailLevel', 'detailLevel', None),
        ('fsSize', 'fsSize', None),
        ('theme', 'theme', DEFAULT_THEME),
        ('showObstacles', 'showObstacles', False),
        ('showTimings', 'showTimings', False),
        ('cursor', 'cursor', 0),
    )

    DETAIL_LEVELS = ['lowest', 'shrunk', 'low', 'default', 'full']

    def __init__(self, *args, **kwargs):
        super(DisplaySettings, self).__init__(*args, **kwargs)
        self.onDetailLevelChanged = Event()
        self.pendingScreenResize = None
        reactor.callLater(0, self.subscribeToScreenResize)

    def subscribeToScreenResize(self):
        # This cannot be done directly during the __init__, because we need to
        # be able to load the display settings in order to initialise the
        # ScreenManager.
        self.app.screenManager.onResize.addListener(self.screenResized)

    def screenResized(self):
        if self.pendingScreenResize:
            self.pendingScreenResize.cancel()
        self.pendingScreenResize = reactor.callLater(
            1, self.processScreenResize)

    def processScreenResize(self):
        self.pendingScreenResize = None
        if self.app.screenManager.isFullScreen():
            return
        self.size = self.app.screenManager.size
        self.save()

    def reset(self):
        SettingsObject.reset(self)

        if self.size is None:
            nWidth, nHeight = max(pygame.display.list_modes())
            self.size = (min(nWidth, 1500), min(nHeight, 850))
        if self.fsSize is None:
            self.fsSize = max(pygame.display.list_modes())
        if self.detailLevel is None:
            if self._old_useAlpha and self._old_windowsTranslucent:
                self.detailLevel = 'full'
            elif self._old_useAlpha or self._old_windowsTranslucent:
                self.detailLevel = 'default'
            else:
                self.detailLevel = 'lowest'
        self.lastDetailLevel = self.detailLevel
        self.applyDetailLevel()
        self.applyCursor()

    def applyDetailLevel(self):
        self.perPixelAlpha = True
        self.alphaOverlays = True
        self.paralaxBackgrounds = True
        self.maxViewportWidth = int(1536 * MAP_TO_SCREEN_SCALE + 0.5)
        self.maxViewportHeight = int(960 * MAP_TO_SCREEN_SCALE + 0.5)
        if self.detailLevel == 'full':
            return

        self.paralaxBackgrounds = False
        if self.detailLevel == 'default':
            return

        self.perPixelAlpha = False
        if self.detailLevel == 'low':
            return

        self.maxViewportWidth = 1024
        self.maxViewportHeight = 768
        if self.detailLevel == 'shrunk':
            return

        # If we're going this low, we really want performance
        self.maxViewportWidth = 800
        self.maxViewportHeight = 600
        self.alphaOverlays = False

    def getSurfaceFlagsHack(self):
        # 2015-10-07 We have empirically determined that setting the flags to
        # pygame.SRCALPHA actually improves the frame rate when compared to
        # setting the flags to 0. Therefore we abandon common sense and just do
        # what works. It could be worth retesting this every so often to make
        # sure it's still the case.

        # return pygame.SRCALPHA if self.perPixelAlpha else 0
        return pygame.SRCALPHA

    def getSize(self):
        if self.fullScreen:
            return self.fsSize
        else:
            return self.size

    def apply(self):
        '''
        Apply the current settings.
        '''
        size = self.getSize()

        # Don't bother changing the screen if the settings that matter haven't
        # changed
        if (size != self.app.screenManager.size) or (
                self.fullScreen != self.app.screenManager.isFullScreen()):
            # Tell the main program to change its screen size.
            self.app.changeScreenSize(size, self.fullScreen)

        if self.lastDetailLevel != self.detailLevel:
            self.lastDetailLevel = self.detailLevel
            self.applyDetailLevel()
            self.onDetailLevelChanged()

        self.applyCursor()

    CURSORS = {
        0: ((32, 32), (15, 15)) + pygame.cursors.compile([
            ' XXXX                      XXXX ',
            'XXXXXX                    XXXXXX',
            'XX...XX                  XX...XX',
            'XX....XX                XX....XX',
            'XX.....XX              XX.....XX',
            ' XX.....XX            XX.....XX ',
            '  XX.....XX          XX.....XX  ',
            '   XX.....XX        XX.....XX   ',
            '    XX.....XX      XX.....XX    ',
            '     XX.....XX    XX.....XX     ',
            '      XX....XX    XX....XX      ',
            '       XX..XXX    XXX..XX       ',
            '        XXXXX      XXXXX        ',
            '         XXX        XXX         ',
            '                                ',
            '                                ',
            '                                ',
            '                                ',
            '         XXX        XXX         ',
            '        XXXXX      XXXXX        ',
            '       XX..XXX    XXX..XX       ',
            '      XX....XX    XX....XX      ',
            '     XX.....XX    XX.....XX     ',
            '    XX.....XX      XX.....XX    ',
            '   XX.....XX        XX.....XX   ',
            '  XX.....XX          XX.....XX  ',
            ' XX.....XX            XX.....XX ',
            'XX.....XX              XX.....XX',
            'XX....XX                XX....XX',
            'XX...XX                  XX...XX',
            'XXXXXX                    XXXXXX',
            ' XXXX                      XXXX ',
        ]),
    }

    def applyCursor(self):
        cursor = self.CURSORS.get(self.cursor, pygame.cursors.broken_x)
        pygame.mouse.set_cursor(*cursor)


class SoundSettings(SettingsObject):
    dataFileName = 'sound'
    attributes = (
        ('soundEnabled', 'playSound', False),
        ('musicEnabled', 'playMusic', True),
        ('musicVolume', 'musicVolume', 100),
        ('soundVolume', 'soundVolume', 100),
    )

    def apply(self):
        '''
        Apply the current settings.
        '''

        if self.musicEnabled != self.app.musicManager.isMusicPlaying():
            if self.musicEnabled:
                self.app.musicManager.playMusic()
            else:
                self.app.musicManager.stopMusic()

        self.app.musicManager.setVolume(self.musicVolume)

        if self.soundEnabled:
            self.app.soundPlayer.setMasterVolume(self.soundVolume / 100.)
        else:
            self.app.soundPlayer.setMasterVolume(0)


class IdentitySettings(SettingsObject):
    dataFileName = 'identity'
    attributes = (
        ('nick', 'nick', None),
        ('usernames', 'usernames', None),
        ('firstTime', 'firstTime', True),
    )

    def __init__(self, app):
        self.app = app
        self.reset()

    def reset(self):
        SettingsObject.reset(self)
        if self.usernames is None:
            self.usernames = {}

    def setNick(self, nick):
        self.nick = nick
        self.save()

    def notFirstTime(self):
        self.firstTime = False
        self.save()


class ConnectionSettings(SettingsObject):
    dataFileName = 'connection'
    attributes = (
        ('servers', 'servers', None),
        ('otherGames', 'otherGames', True),
        ('lanGames', 'lanGames', 'afterInet'),
        ('createGames', 'createGames', True),
    )

    def __init__(self, app):
        SettingsObject.__init__(self, app)
        if self.servers is None:
            self.servers = [('localhost', 6787, 'http://localhost:8080/')]
            if trosnoth.version.release:
                self.servers.append(
                    ('play.trosnoth.org', 6787, 'http://play.trosnoth.org/'))
        else:
            # For developers who have saved settings.
            for i, s in enumerate(self.servers):
                if s == ('localhost', 6787):
                    self.servers[i] = (
                        'localhost', 6787, 'http://localhost:8080/')
                elif len(s) == 2:
                    self.servers[i] = (s[0], s[1], 'http://%s/' % (s[0],))


class AuthServerSettings(SettingsObject):
    '''
    Stores general settings for the authentication server.
    '''
    attributes = (
        ('keyLength', 'keyLength', 512),
        ('lobbySize', 'lobbySize', (2, 1)),
        ('maxPerTeam', 'playersPerTeam', 5),
        ('allowNewUsers', 'allowNewUsers', True),
        ('privateMsg', 'privateMsg', None),
        ('elephantOwners', 'elephantOwners', ()),
        ('serverName', 'serverName', 'Trosnoth server'),
        ('homeUrl', 'homeUrl', None),
        ('hostName', 'hostName', None),
        ('xmppUsername', 'xmppUsername', None),
        ('xmppPassword', 'xmppPassword', None),
    )

    def __init__(self, dataPath):
        self.dataFileName = os.path.join(dataPath, 'settings')
        self.reset()

    def _getSettingsFilename(self):
        return self.dataFileName

    def reset(self):
        SettingsObject.reset(self)
        self.elephantOwners = set(p.lower() for p in self.elephantOwners)
        if self.privateMsg is None or len(self.privateMsg) == 0:
            self.privateMsg = 'This is a private server.'

    def createNotificationClient(self):
        '''
        Either returns an xmpp.NotificationClient or None.
        '''
        username = self.xmppUsername
        password = self.xmppPassword
        if (not username) or (not password):
            return None

        try:
            from trosnoth.network.xmpp import NotificationClient
        except ImportError, E:
            log.warning('Could not import xmpp notification client: %s', E)
            return None

        return NotificationClient(username, password)


class TournamentSettings(SettingsObject):
    '''
    Stores general settings for tournaments on the authentication server.
    '''
    attributes = (
        ('teamNames', 'teamNames', ()),
        ('minTeamSize', 'minTeamSize', 7),
        ('maxTeamSize', 'maxTeamSize', 8),
        ('mapSize', 'mapSize', (3, 2)),
        ('gameDuration', 'gameDuration', 2700),
    )

    def __init__(self, dataPath):
        self.dataFileName = os.path.join(dataPath, 'tournaments')
        self.reset()

    def _getSettingsFilename(self):
        return self.dataFileName



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2013 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import __builtin__

from rlcompleter import Completer

from trosnoth.utils.twist import WeakCallLater

from twisted.conch import manhole


class Manhole(manhole.ColoredManhole):

    tabCount = 0

    def sayHello(self):
        self.terminal.reset()
        self.terminal.write("\x1b[97mWelcome to \x1b[36mTrosnoth!")
        self.terminal.nextLine()
        helper = self.namespace['helper']
        for line in helper.getBanner():
            self.terminal.write(line)
            self.terminal.nextLine()
        self.terminal.write('>>> ')

    def connectionMade(self):
        super(Manhole, self).connectionMade()
        WeakCallLater(0.1, self, 'sayHello')

    def keystrokeReceived(self, keyID, modifier):
        super(Manhole, self).keystrokeReceived(keyID, modifier)
        self.tabCount += 1 if keyID == "\t" else 0

    def handle_TAB(self):
        text = "".join(self.lineBuffer).split(' ')[-1]
        if len(text) == 0:
            # Bell character
            self.terminal.write("\a")
            return

        completer = Completer(self.namespace)

        if completer.complete(text, 0):
            allMatches = list(set(completer.matches))

            # Get rid of a bunch of cruft
            builtins = __builtin__.keys()
            matches = [x for x in allMatches
                       if x.strip('(') not in builtins and "__" not in x]
            matches.sort()

            # If there are no matches, ring the terminal bell
            # If there's only one match, autocomplete it
            # If there's more than one match, print a list of matches
            if len(matches) == 0:
                self.terminal.write("\a")
                return
            elif len(matches) == 1:
                length = len(text)
                self.lineBuffer = self.lineBuffer[:-length]
                self.lineBuffer.extend(matches[0])
                self.lineBufferIndex = len(self.lineBuffer)
            else:
                # Remove text before the last dot, for brevity
                if "." in matches[0]:
                    matches = [x[x.rfind(".") + 1:] for x in matches]
                self.terminal.nextLine()
                self.terminal.write(repr(matches))
                self.terminal.nextLine()
                self.terminal.write("%s%s" % (self.ps[self.pn], ""))

            self.terminal.eraseLine()
            self.terminal.cursorBackward(self.lineBufferIndex + 5)
            self.terminal.write(
                "%s%s" % (self.ps[self.pn], "".join(self.lineBuffer)))

        else:
            self.terminal.write("\a")



import pygame

from trosnoth.gui.framework.elements import TextButton
from trosnoth.gui.common import (Location, ScaledSize,
        ScaledScreenAttachedPoint)


def mainButton(app, text, onClick, pos, anchor='topleft', hugeFont=False,
        smallFont=False):
    pos = Location(ScaledScreenAttachedPoint(ScaledSize(pos[0], pos[1]),
            'topleft'), anchor)
    if hugeFont:
        font = app.screenManager.fonts.hugeMenuFont
    elif smallFont:
        font = app.screenManager.fonts.menuFont
    else:
        font = app.screenManager.fonts.mainMenuFont
    result = TextButton(app, pos, text, font, app.theme.colours.mainMenuColour,
            app.theme.colours.mainMenuHighlight)
    result.onClick.addListener(lambda sender: onClick())
    return result


def button(app, text, onClick, pos, anchor='topleft', hugeFont=False,
        smallFont=False, firstColour=None, secondColour=None):
    if firstColour is None:
        firstColour = app.theme.colours.secondMenuColour
    if secondColour is None:
        secondColour = app.theme.colours.mainMenuHighlight
    pos = Location(ScaledScreenAttachedPoint(ScaledSize(pos[0], pos[1]),
            anchor), anchor)
    if hugeFont:
        font = app.screenManager.fonts.hugeMenuFont
    elif smallFont:
        font = app.screenManager.fonts.menuFont
    else:
        font = app.screenManager.fonts.mainMenuFont
    result = TextButton(app, pos, text, font, firstColour, secondColour)
    result.onClick.addListener(lambda sender: onClick())
    return result


def setAlpha(surface, alpha, alphaSurface=None):
    '''
    Sets the alpha of the given surface, even if it already has per-pixel
    alpha.
    Note that if per-pixel alphas are enabled, this will actually change the
    transparency of the pixels in the image, so calling it again with alpha=255
    will not reverse it.
    '''
    if surface.get_flags() & pygame.SRCALPHA:
        if alpha >= 255:
            return
        if not alphaSurface:
            alphaSurface = pygame.Surface(
                surface.get_rect().size, pygame.SRCALPHA)

        alphaSurface.fill((255, 255, 255, alpha))
        surface.blit(
            alphaSurface, (0, 0), special_flags=pygame.BLEND_RGBA_MULT)
    else:
        if alpha >= 255:
            alpha = None
        surface.set_alpha(alpha)



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from trosnoth.gui.framework import framework, elements, scrollingText
from trosnoth.gui.common import (ScaledArea, Location, ScaledSize,
        ScaledScreenAttachedPoint)
from trosnoth.utils.event import Event

from trosnoth.data import getPath
import trosnoth.data.startupMenu as startupMenu

class ScalingSpeed(object):
    def __init__(self, baseSpeed):
        self.baseSpeed = baseSpeed

    def getSpeed(self, app):
        return self.baseSpeed * app.screenManager.scaleFactor

class CreditsScreen(framework.CompoundElement):
    def __init__(self, app, colour, onCancel=None, speed=None, loop=True,
            startOff=False, backText='back to main menu',
            highlight=(255, 255, 255)):
        super(CreditsScreen, self).__init__(app)
        self.colour = colour
        self.highlight = highlight

        self.onCancel = Event()
        if onCancel is not None:
            self.onCancel.addListener(onCancel)

        f = open(getPath(startupMenu, 'credits.txt'), 'rU')
        area = ScaledArea(50,130,900, 540)
        fonts = {'body': self.app.fonts.creditsFont,
         'h1': self.app.fonts.creditsH1,
         'h2': self.app.fonts.creditsH2}
        text = f.read()
        if not startOff:
            # Make at least some room on top and bottom
            text = '\n'*3 + text + '\n'*2
        self.credits = scrollingText.ScrollingText(self.app,
                        area, text, self.colour,
                        fonts, textAlign='middle', loop=loop,
                        startOff=startOff)
        self.credits.setShadowColour((192, 192, 192))

        self.credits.setAutoscroll(True)
        if speed is None:
            speed = 80
        self.credits.setSpeed(ScalingSpeed(speed))
        self.credits.setBorder(False)

        self.backText = backText
        cancelButton = self.button(backText, self.onCancel.execute,
                (-50, -30), 'bottomright')

        self.elements = [self.credits, cancelButton]

    @property
    def finished(self):
        return self.credits.reachedEnd

    def setColour(self, colour):
        self.colour = colour
        self.elements[-1] = self.button(self.backText,
                self.onCancel.execute, (-50, -30), 'bottomright')
        self.credits.setColour(colour)

    def button(self, text, onClick, pos, anchor='topleft', hugeFont=False):
        pos = Location(ScaledScreenAttachedPoint(ScaledSize(pos[0], pos[1]),
                                                 anchor), anchor)
        if hugeFont:
            font = self.app.screenManager.fonts.hugeMenuFont
        else:
            font = self.app.screenManager.fonts.bigMenuFont
        result = elements.TextButton(self.app, pos, text, font, self.colour,
                                     self.highlight)
        result.onClick.addListener(lambda sender: onClick())
        return result

    def restart(self):
        self.credits.returnToTop()



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from trosnoth.messages.base import TICK_LIMIT   # noqa
from trosnoth.messages.gameplay import (    # noqa
    TickMsg, TaggingZoneMsg, CreateCollectableStarMsg,
    RemoveCollectableStarMsg, PlayerUpdateMsg, RespawnMsg, CheckSyncMsg,
    RespawnRequestMsg, CannotRespawnMsg, DelayUpdatedMsg,
    ResyncPlayerMsg, ResyncAcknowledgedMsg, ShotHitPlayerMsg,
    UpdatePlayerStateMsg, AimPlayerAtMsg, ShootMsg, ShotFiredMsg,
    FireShoxwaveMsg, PlayerNoticedZombieHitMsg, ChatFromServerMsg, ChatMsg,
)
from trosnoth.messages.setup import (       # noqa
    ChangeNicknameMsg, PlayerIsReadyMsg, SetPreferredTeamMsg,
    PreferredTeamSelectedMsg, SetPreferredSizeMsg, SetPreferredDurationMsg,
    GameStartMsg, GameOverMsg, ReturnToLobbyMsg, SetGameModeMsg,
    SetGameSpeedMsg, SetTeamNameMsg, StartingSoonMsg, ChangeTimeLimitMsg,
    AddPlayerMsg, SetPlayerTeamMsg, RemovePlayerMsg, WorldLoadingMsg,
    JoinRequestMsg, CannotJoinMsg, SetAgentPlayerMsg, InitClientMsg,
    ConnectionLostMsg, WorldResetMsg, ZoneStateMsg,
)
from trosnoth.messages.special import (     # noqa
    PlayerHasElephantMsg, PlayerHasTrosballMsg, TrosballPositionMsg,
    ThrowTrosballMsg, TrosballTagMsg, AchievementUnlockedMsg,
)
from trosnoth.messages.upgrades import (    # noqa
    AbandonUpgradeMsg, BuyUpgradeMsg, PlayerHasUpgradeMsg, PlayerStarsSpentMsg,
    AwardPlayerStarMsg, CannotBuyUpgradeMsg,
    UpgradeChangedMsg, UpgradeApprovedMsg,
)



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from framework import Element
import pygame

class StatList(Element):
    '''
    @param items A list of items to retrieve text from
    @param function A callable object which will be run on each item
    '''
    def __init__(self, app, pos, items, font,
            colourFunction=lambda item: ((0,0,0)), function=str, align='left'):
        super(StatList, self).__init__(app)
        self.pos = pos
        self.items = items
        self.font = font
        self.align = align
        self.colourFunction = colourFunction
        self.function = function
        self._visible = False

    def getVisibility(self):
        return self._visible

    def setVisible(self, isVisible = True):
        self._visible = isVisible

    def _moveRect(self, rect):
        if hasattr(self.pos, 'apply'):
            self.pos.apply(self.app, rect)
            return rect.topleft
        else:
            return self.pos

    def draw(self, screen):
        '''Gives the element a chance to draw itself onto the screen.'''
        if self._visible:
            totalHeight = 0
            maxWidth = 0
            surfaces = []
            colours = []
            for i in range(0, len(self.items)):
                item = self.items[i]
                text = self.function(item)
                colour = self.colourFunction(item)
                colours.append(colour)
                s = self.font.render(self.app, text, True, colour)
                maxWidth = max(maxWidth, s.get_width())
                totalHeight += s.get_height()
                surfaces.append(s)

            tempRect = pygame.Rect((0,0),(maxWidth, totalHeight))
            self._moveRect(tempRect)
            currentHeight = tempRect.top
            for s in surfaces:
                r = s.get_rect()
                setattr(r, self.align, getattr(tempRect, self.align))
                r.top = currentHeight
                screen.blit(s, r.topleft)
                currentHeight += s.get_height()



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from trosnoth.gui.app import Application
from trosnoth.gui.framework import framework, elements, scrollingText
from trosnoth.gui.fonts.font import Font
import pygame
from random import randint, shuffle

class Interface(framework.CompoundElement):
    def __init__(self, app):
        super(Interface, self).__init__(app)
        backdrop = pygame.Surface(app.screenManager.scaledSize)
        backdrop.fill((128,0,128))
        f = open('credits.txt', 'r')
        credPos = (50,50)
        size = (500, 350)
        fonts = {'body': self.app.fonts.creditsFont,
         'h1': self.app.fonts.creditsH1,
         'h2': self.app.fonts.creditsH2}
        cred = scrollingText.ScrollingText(self.app,
                        credPos, size, f.read(), (0,0,0),
                        fonts, bgColour = (255,255,255), autoScroll = False,
                        speed = 100, align = 'middle', border=True,
                        loop = True)
        def doIt():
            print("Done Scrolling")
        cred.onFinishedScrolling.addListener(lambda: doIt())
        
        self.elements = [elements.PictureElement(app, backdrop), cred]

size = (600,450)

if __name__ == '__main__':
    a = Application(size, 0, "Testing", Interface)
    a.run()



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import logging

from trosnoth.messages import AwardPlayerStarMsg, RemoveCollectableStarMsg
from trosnoth.model.unit import Bouncy, CollectableUnit

log = logging.getLogger('star')


class CollectableStar(Bouncy, CollectableUnit):
    # The following values control star movement.
    maxFallVel = 540            # pix/s
    gravity = 1000              # pix/s/s
    initVel = 480

    HALF_WIDTH = 5
    HALF_HEIGHT = 5

    def __init__(self, world, id, pos, *args, **kwargs):
        super(CollectableStar, self).__init__(world, *args, **kwargs)
        self.id = id
        self.creationTime = self.world.timer.getElapsedTime()
        self.vanished = False

        # Place myself.
        self.pos = pos
        self.xVel = 0
        self.yVel = 0

    def collidedWithPlayer(self, player):
        self.world.sendServerCommand(AwardPlayerStarMsg(player.id))

        if self.vanished:
            # It's already been removed from the main set of stars.
            self.world.deadStars.remove(self)
        else:
            del self.world.collectableStars[self.id]
            self.world.sendServerCommand(RemoveCollectableStarMsg(self.id))

    def removeDueToTime(self):
        self.vanished = True
        self.world.deadStars.add(self)
        del self.world.collectableStars[self.id]
        self.world.sendServerCommand(RemoveCollectableStarMsg(self.id))

    def getGravity(self):
        return self.gravity

    def getMaxFallVel(self):
        return self.maxFallVel



import pygame
import trosnoth.gui.framework.framework as framework
from trosnoth.utils.event import Event
from trosnoth.gui.common import ScaledArea

class Slider(framework.Element):
    def __init__(self, app, area, bounds=None, snap=False):
        super(Slider, self).__init__(app)
        self.app = app
        self.area = area
        self.beingDragged = False
        self.snap = snap
        if not bounds:
            bounds = 0, 100
        self.lowBound, self.highBound = bounds
        self.unscaledSlider = 20
        self.sliderRectWidth = int(self.unscaledSlider *
                app.screenManager.scaleFactor)
        self.lineColour = (0,192,0)
        self.sliderColour = (0,255,0)
        self.onValueChanged = Event()
        self.onSlide = Event()

        self.largeArea = ScaledArea(self.area.point.val[0] +
                self.unscaledSlider / 2, self.area.point.val[1],
                self.area.size.val[0] - self.unscaledSlider + 1,
                self.area.size.val[1])

        self.area, self.largeArea = self.largeArea, self.area

        self.sliderVal = self.highBound

    def doSnap(self, value):
        if self.snap:
            return int(value + 0.5)
        return value

    def getVal(self):
        return self.sliderVal

    def setVal(self, val):
        old = self.sliderVal
        self.sliderVal = self.doSnap(val)
        if old != self.sliderVal:
            self.onValueChanged.execute(self.sliderVal)

    def setRange(self, low, high):
        self.lowBound = low
        self.highBound = high
        if self.sliderVal > self.highBound:
            self.sliderVal = self.highBound
        elif self.sliderVal < self.lowBound:
            self.sliderVal = self.lowBound

    def setSliderColour(self, colour):
        self.sliderColour = colour

    def setLineColour(self, colour):
        self.lineColour = colour

    def _getRange(self):
        return self.highBound - self.lowBound

    def _getRect(self):
        return self.area.getRect(self.app)

    def _getPt(self):
        return self._getRect().topleft

    def _getSize(self):
        return self._getRect().size

    def _valPerPixel(self):
        return self._getSize()[0] / (self._getRange() + 0.)

    def _pixelPerVal(self):
        return self._getRange() / (self._getSize()[0] + 0.)

    # Number of pixels from 0 to width
    def _relPx(self, x):
        return min(max(0, (x - self._getPt()[0])), self._getSize()[0])

    def _mouseAt(self, x):
        old = self.sliderVal
        self.sliderVal = self.lowBound + self._pixelPerVal() * self._relPx(x)
        if old != self.sliderVal:
            self.onSlide.execute(self.sliderVal)

    def _getSliderRect(self):
        # Ignore pos for now
        sliderRect = pygame.Rect(0,0, self._getSliderRectWidth(),
                self._getSize()[1])
        pt = self._getPt()
        sliderRect.midtop = ((self.sliderVal - self.lowBound) *
                self._valPerPixel() + pt[0], pt[1])
        return sliderRect

    def _getLineWidth(self):
        return max(1, self._getSize()[1] / 8)

    def _getSliderRectWidth(self):
        return int(self.unscaledSlider * self.app.screenManager.scaleFactor)


    def processEvent(self, event):
        if (event.type == pygame.MOUSEBUTTONDOWN and
                self.largeArea.getRect(self.app).collidepoint(event.pos)):
            # Each click of the scroll wheel moves the slider by 5%
            percent = int((self.highBound - self.lowBound) * 0.05)

            if event.button == 1:
                self.beingDragged = True
                self._mouseAt(event.pos[0])
            elif event.button == 4:
                self.setVal(min(self.highBound,
                                self.sliderVal + percent))
            elif event.button == 5:
                self.setVal(max(self.lowBound,
                                self.sliderVal - percent))
            else:
                return event

        elif event.type == pygame.MOUSEMOTION and self.beingDragged:
            self._mouseAt(event.pos[0])

        elif (event.type == pygame.MOUSEBUTTONUP and event.button == 1 and
                self.beingDragged):
            self.beingDragged = False
            self.sliderVal = self.doSnap(self.sliderVal)
            self.onValueChanged.execute(self.sliderVal)
        else:
            return event
        return None

    def draw(self, surface):
        r = self.largeArea.getRect(self.app)
        lineWidth = self._getLineWidth()
        pygame.draw.line(
            surface, self.lineColour, r.midleft, r.midright, lineWidth)

        if self.snap:
            low = r.left
            high = r.right
            count = self.highBound - self.lowBound + 0.
            y0 = r.centery - 0.25 * r.height
            y1 = r.centery + 0.25 * r.height
            for i in xrange(self.lowBound, self.highBound + 1):
                x = (high * i + low * (count - i)) / count
                pygame.draw.line(
                    surface, self.lineColour, (x, y0), (x, y1), lineWidth)

        surface.fill(self.sliderColour, self._getSliderRect())




# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import pygame

from trosnoth.utils.event import Event
from trosnoth.gui.common import (uniqueColour, AttachedPoint, Location,
        TextImage)
from framework import Element
from elements import HoverButton

class ListBox(Element):
    def __init__(self, app, area, items, font, colour, hlColour=None,
                 showBtns=True):
        super(ListBox, self).__init__(app)
        self.area = area
        self.items = items
        self.font = font
        self.showBtns = showBtns
        self.onValueChanged = Event()

        self.setColour(colour, hlColour)

        self.offset = 0
        self.index = -1


        # Create up and down buttons.
        if showBtns:
            img1 = TextImage('up', font, colour)
            img2 = TextImage('up', font, self.hlColour)
            self.upBtn = HoverButton(self.app, Location(AttachedPoint((0,0),
                    self._getRect, 'topright'), 'topright'), img1, img2)
            self.upBtn.onClick.addListener(self.upClick)
            img1 = TextImage('down', font, colour)
            img2 = TextImage('down', font, self.hlColour)
            self.dnBtn = HoverButton(self.app, Location(AttachedPoint((0,0),
                    self._getRect, 'bottomright'), 'bottomright'), img1, img2)
            self.dnBtn.onClick.addListener(self.dnClick)

    def _getRect(self):
        return self.area.getRect(self.app)

    def _getMaxInView(self):
        return int((self._getSize()[1] - 1) / self._getItemHeight()) + 1

    def _getItemHeight(self):
        return int(self.font.getLineSize(self.app))

    def setColour(self, colour, hlColour=None):
        self.colour = colour
        if hlColour:
            self.hlColour = hlColour
        else:
            self.hlColour = tuple((255 + i) / 2 for i in colour)

        self.bgColour = uniqueColour([self.hlColour, self.colour])

    def setItems(self, items):
        self.items = items
        if self.index >= len(self.items):
            self.index = -1

    def getItem(self, index=None):
        if index == None:
            index = self.index
        if index >= 0 and index < len(self.items):
            return self.items[index]
        raise ValueError, "Listbox doesn't have that many items"

    def getNumItems(self):
        return len(self.items)

    def setIndex(self, index):
        if index >= self.getNumItems():
            raise ValueError, "Listbox doesn't have that many items"
        self.index = index

    def getIndex(self):
        return self.index

    def _getSize(self):
        return self._getRect().size

    def _getPt(self):
        rect = self._getRect()
        return rect.topleft


    def draw(self, screen):
        rect = self._getRect()
        pos = rect.topleft


        self.offset = min(self.offset, max(0, len(self.items) -
                                           self._getMaxInView() + 1))

        # Draw the items.
        y = pos[1]
        for i in xrange(self.offset, min(self.offset + self._getMaxInView(),
                len(self.items))):
            text = self.font.render(self.app, self.items[i], True, self.colour)
            self._blitTextToScreen(screen, text, (pos[0], y))

            # Highlight selected one.
            if i == self.index:
                text = self.font.render(self.app, self.items[i], True,
                        self.hlColour)
                hlOffset = int(self._getItemHeight() / 30. + .5)
                self._blitTextToScreen(screen, text,
                        (pos[0] + hlOffset, y + hlOffset))


            y = y + self._getItemHeight()


        if self.showBtns:
            self.upBtn.draw(screen)
            self.dnBtn.draw(screen)

    def _blitTextToScreen(self, screen, textImage, placement):
        rect = self._getRect()
        textRect = textImage.get_rect()
        textRect.topleft = placement
        clippedRect = textRect.clip(rect)
        clippedRect.topleft = (0,0)

        screen.blit(textImage, placement, clippedRect)

    def processEvent(self, event):
        if self.showBtns:
            event = self.upBtn.processEvent(event)
            if not event: return
            event = self.dnBtn.processEvent(event)
            if not event: return

        # Click on element selects it.
        if (event.type == pygame.MOUSEBUTTONDOWN and
                self._getRect().collidepoint(event.pos)):
            # Left-click
            if event.button == 1:
                rect = self._getRect()
                pos = rect.topleft
                index = ((event.pos[1] - pos[1]) / self._getItemHeight() +
                        self.offset)
                if index < len(self.items):
                    self.index = index
                    self.onValueChanged.execute(index)
                else:
                    return event
            # Scroll up
            elif event.button == 4:
                self._changeSelection(max(self.index - 1, 0))
            # Scroll down
            elif event.button == 5:
                self._changeSelection(min(self.index + 1, len(self.items) - 1))
            else:
                return event

        elif event.type == pygame.KEYDOWN:
            if event.key == pygame.K_UP:
                self._changeSelection(max(self.index - 1, 0))
            elif event.key == pygame.K_DOWN:
                self._changeSelection(min(self.index + 1, len(self.items) - 1))
            elif event.key == pygame.K_PAGEUP:
                self._changeSelection(max(self.index - self._getMaxInView(), 0))
            elif event.key == pygame.K_PAGEDOWN:
                self._changeSelection(min(self.index + self._getMaxInView(),
                        len(self.items) - 1))
            elif event.key == pygame.K_HOME:
                self._changeSelection(0)
            elif event.key == pygame.K_END:
                self._changeSelection(len(self.items) - 1)
            else:
                return event
        else:
            return event
        return None

    def _changeSelection(self, newIndex):
        self.setIndex(newIndex)
        self._putSelectedInView()
        self.onValueChanged.execute(self.index)

    def _putSelectedInView(self):
        if self.offset+ self._getMaxInView() <= self.index:
            self.offset = (min(self.index + 3, len(self.items)) -
                    self._getMaxInView())
        elif self.offset > self.index:
            self.offset = max(self.index - 2, 0)

    def tick(self, deltaT):
        if not self.showBtns: return

        self.upBtn.tick(deltaT)
        self.dnBtn.tick(deltaT)

    def upClick(self, btn):
        self.offset = max(self.offset - 1, 0)

    def dnClick(self, btn):
        self.offset = self.offset + 1




from trosnoth.gui.framework.dialogbox import (
    DialogBox, DialogResult, DialogBoxAttachedPoint,
)
from trosnoth.gui.common import ScaledSize, Area, ScaledLocation, Location
from trosnoth.gui.framework import elements, prompt


SpectateResult = object()


class JoinGameDialog(DialogBox):
    def __init__(self, app, controller, world):
        super(JoinGameDialog, self).__init__(
            app, ScaledSize(512, 314), 'Join Game')
        self.result = None
        self.controller = controller
        self.selectedTeam = None

        fonts = self.app.screenManager.fonts
        self.nickBox = prompt.InputBox(
            self.app,
            Area(
                DialogBoxAttachedPoint(self, ScaledSize(0, 40), 'midtop'),
                ScaledSize(200, 60), 'midtop'),
            '',
            font=fonts.menuFont,
            maxLength=30,
        )
        self.nickBox.onClick.addListener(self.setFocus)
        self.nickBox.onTab.addListener(lambda sender: self.clearFocus())
        name = app.identitySettings.nick
        if name is not None:
            self.nickBox.setValue(name)

        colours = app.theme.colours
        self.cantJoinYet = elements.TextElement(
            self.app,
            '',
            fonts.ingameMenuFont,
            ScaledLocation(256, 115, 'center'),
            colours.cannotJoinColour,
        )

        teamA = world.teams[0]
        teamB = world.teams[1]

        self.elements = [
            elements.TextElement(
                self.app,
                'Please enter your nick:',
                fonts.smallMenuFont,
                Location(
                    DialogBoxAttachedPoint(self, ScaledSize(0, 10), 'midtop'),
                    'midtop'),
                colours.black,
            ),
            self.nickBox,
            self.cantJoinYet,
            elements.TextElement(
                self.app,
                'Select team:',
                fonts.smallMenuFont,
                Location(
                    DialogBoxAttachedPoint(self, ScaledSize(0, 130), 'midtop'),
                    'midtop'),
                colours.black,
            ),

            elements.TextButton(
                self.app,
                Location(
                    DialogBoxAttachedPoint(
                        self, ScaledSize(-25, 160), 'midtop'),
                    'topright'),
                str(teamA),
                fonts.menuFont,
                colours.team1msg,
                colours.white,
                onClick=lambda obj: self.joinTeam(teamA)
            ),
            elements.TextButton(
                self.app,
                Location(
                    DialogBoxAttachedPoint(
                        self, ScaledSize(25, 160), 'midtop'),
                    'topleft'),
                str(teamB),
                fonts.menuFont,
                colours.team2msg,
                colours.white,
                onClick=lambda obj: self.joinTeam(teamB)
            ),
            elements.TextButton(
                self.app,
                Location(
                    DialogBoxAttachedPoint(
                        self, ScaledSize(-25, 210), 'midtop'),
                    'topright'),
                'Automatic',
                fonts.menuFont,
                colours.inGameButtonColour,
                colours.white,
                onClick=lambda obj: self.joinTeam()
            ),
            elements.TextButton(
                self.app,
                Location(
                    DialogBoxAttachedPoint(
                        self, ScaledSize(25, 210), 'midtop'),
                    'topleft'),
                'Spectator',
                fonts.menuFont,
                colours.inGameButtonColour,
                colours.white,
                onClick=lambda obj: self.spectate()
            ),

            elements.TextButton(
                self.app,
                Location(
                    DialogBoxAttachedPoint(
                        self, ScaledSize(0, -10), 'midbottom'),
                    'midbottom'),
                'Cancel',
                fonts.menuFont,
                colours.inGameButtonColour,
                colours.white,
                onClick=self.cancel
            )
        ]
        self.setColours(
            colours.joinGameBorderColour,
            colours.joinGameTitleColour,
            colours.joinGameBackgroundColour)
        self.setFocus(self.nickBox)

    def joinTeam(self, team=None):
        self.selectedTeam = team
        self.cantJoinYet.setText('')

        nick = self.nickBox.value
        if nick == '' or nick.isspace():
            # Disallow all-whitespace nicks
            return

        self.result = DialogResult.OK
        self.close()

    def spectate(self):
        self.selectedTeam = None
        self.cantJoinYet.setText('')

        self.result = SpectateResult
        self.close()

    def cancel(self, sender):
        self.result = DialogResult.Cancel
        self.close()


class JoiningDialog(DialogBox):
    def __init__(self, app, controller):
        super(JoiningDialog, self).__init__(
            app, ScaledSize(530, 180), 'Trosnoth')
        colours = app.theme.colours
        self.controller = controller

        fonts = self.app.screenManager.fonts
        self.text = elements.TextElement(
            self.app,
            '',
            fonts.menuFont,
            Location(
                DialogBoxAttachedPoint(self, ScaledSize(0, 40), 'midtop'),
                'midtop'),
            colour=colours.joiningColour,
        )

        self.elements = [
            self.text,
            elements.TextButton(
                self.app,
                Location(
                    DialogBoxAttachedPoint(
                        self, ScaledSize(0, -10), 'midbottom'),
                    'midbottom'),
                'Cancel',
                fonts.menuFont,
                colours.inGameButtonColour,
                colours.white,
                onClick=controller.cancelJoin,
            ),
        ]
        self.setColours(
            colours.joinGameBorderColour,
            colours.joinGameTitleColour,
            colours.joinGameBackgroundColour)

    def show(self, nick):
        self.text.setText('Joining as %s...' % (nick,))
        DialogBox.show(self)
